#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The agri_l1 reader tests package."""

import os
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr

from satpy.tests.reader_tests.test_hdf5_utils import FakeHDF5FileHandler

ALL_BAND_NAMES = ["C01", "C02", "C03", "C04", "C05", "C06", "C07"]
RESOLUTION_LIST = [250, 500, 2000]

CHANNELS_BY_RESOLUTION = {250: ["C01"],
                          500: ["C01", "C02", "C03", "C04", "C05", "C06"],
                          2000: ALL_BAND_NAMES,
                          "GEO": "solar_azimuth_angle"
                          }

AREA_EXTENTS_BY_RESOLUTION = {
    250: (896278.676104, 562456.016066, 895155.242397, 452480.774883),
    500: (896153.676104, 562331.016066, 895155.242397, 452480.774883),
    2000: (895403.676104, 561581.016066, 895155.242397, 452480.774883)
}


class FakeHDF5FileHandler2(FakeHDF5FileHandler):
    """Swap-in HDF5 File Handler."""

    def make_test_data(self, cwl, ch, prefix, dims, file_type):
        """Make test data."""
        if prefix == "CAL":
            data = xr.DataArray(
                da.from_array((np.arange(10.) + 1.) / 10., [dims[0] * dims[1]]),
                attrs={
                    "Slope": np.array(1.), "Intercept": np.array(0.),
                    "FillValue": np.array(-65535.0),
                    "units": "NUL",
                    "center_wavelength": "{}um".format(cwl).encode("utf-8"),
                    "band_names": "band{}(band number is range from 1 to 14)"
                    .format(ch).encode("utf-8"),
                    "long_name": "Calibration table of {}um Channel".format(cwl).encode("utf-8"),
                    "valid_range": np.array([0, 1.5]),
                },
                dims="_const")

        elif prefix == "NOM":
            data = xr.DataArray(
                da.from_array(np.arange(10, dtype=np.uint16).reshape((2, 5)) + 1,
                              [dim for dim in dims]),
                attrs={
                    "Slope": np.array(1.), "Intercept": np.array(0.),
                    "FillValue": np.array(65535),
                    "units": "DN",
                    "center_wavelength": "{}um".format(cwl).encode("utf-8"),
                    "band_names": "band{}(band number is range from 1 to 7)"
                    .format(ch).encode("utf-8"),
                    "long_name": "Calibration table of {}um Channel".format(cwl).encode("utf-8"),
                    "valid_range": np.array([0, 4095]),
                },
                dims=("_RegLength", "_RegWidth"))

        elif prefix == "GEO":
            data = xr.DataArray(
                da.from_array(np.arange(10, dtype=np.float32).reshape((2, 5)) + 1,
                              [dim for dim in dims]),
                attrs={
                    "Slope": np.array(1.), "Intercept": np.array(0.),
                    "FillValue": np.array(65535.),
                    "units": "NUL",
                    "band_names": "NUL",
                    "valid_range": np.array([0., 360.]),
                },
                dims=("_RegLength", "_RegWidth"))

        elif prefix == "COEF":
            if file_type == "250":
                data = self._create_coeff_array(1)

            elif file_type == "500":
                data = self._create_coeff_array(6)

            elif file_type == "2000":
                data = self._create_coeff_array(7)

        return data

    def _create_coeff_array(self, nb_channels):
        data = xr.DataArray(
            da.from_array((np.arange(nb_channels * 2).reshape((nb_channels, 2)) + 1.) /
                          np.array([1E4, 1E2]), [nb_channels, 2]),
            attrs={
                "Slope": 1., "Intercept": 0.,
                "FillValue": 0,
                "units": "NUL",
                "band_names": "NUL",
                "long_name": b"Calibration coefficient (SCALE and OFFSET)",
                "valid_range": [-500, 500],
            },
            dims=("_num_channel", "_coefs"))
        return data

    def _create_channel_data(self, chs, cwls, file_type):
        dim_0 = 2
        dim_1 = 5
        data = {}
        for index, _cwl in enumerate(cwls):
            data["Calibration/CALChannel" + "%02d" % chs[index]] = self.make_test_data(cwls[index], chs[index], "CAL",
                                                                                       [dim_0, dim_1], file_type)
            data["Data/NOMChannel" + "%02d" % chs[index]] = self.make_test_data(cwls[index], chs[index], "NOM",
                                                                                [dim_0, dim_1], file_type)
            data["Calibration/CALIBRATION_COEF(SCALE+OFFSET)"] = self.make_test_data(cwls[index], chs[index], "COEF",
                                                                                     [dim_0, dim_1], file_type)
        return data

    def _get_250m_data(self, file_type):
        chs = [1]
        cwls = [0.675]
        data = self._create_channel_data(chs, cwls, file_type)

        return data

    def _get_500m_data(self, file_type):
        chs = [1, 2, 3, 4, 5, 6]
        cwls = [0.675, 0.47, 0.545, 0.645, 1.378, 1.61]
        data = self._create_channel_data(chs, cwls, file_type)

        return data

    def _get_2km_data(self, file_type):
        chs = [1, 2, 3, 4, 5, 6, 7]
        cwls = [0.675, 0.47, 0.545, 0.645, 1.378, 1.61, 11.4]
        data = self._create_channel_data(chs, cwls, file_type)

        return data

    def _get_geo_data(self, file_type):
        dim_0 = 2
        dim_1 = 5
        data = {"Navigation/NOMSunAzimuth": self.make_test_data("NUL", "NUL", "GEO",
                                                                [dim_0, dim_1], file_type)}
        return data

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        global_attrs = {
            "/attr/NOMSubSatLat": np.array(0.0),
            "/attr/NOMSubSatLon": np.array(133.0),
            "/attr/NOMSatHeight": np.array(3.5786E7),
            "/attr/Semi_major_axis": np.array(6378.14),
            "/attr/Semi_minor_axis": np.array(6353.28),
            "/attr/OBIType": "REGX",
            "/attr/RegLength": np.array(2.0),
            "/attr/RegWidth": np.array(5.0),
            "/attr/Corner-Point Latitudes": np.array((4.1, 5.1, 4.1, 5.1)),
            "/attr/Corner-Point Longitudes": np.array((141.1, 141.1, 141.1, 151.1)),
            "/attr/Begin Line Number": np.array(0),
            "/attr/End Line Number": np.array(1),
            "/attr/Observing Beginning Date": "2019-06-03", "/attr/Observing Beginning Time": "00:30:01.807",
            "/attr/Observing Ending Date": "2019-06-03", "/attr/Observing Ending Time": "00:34:07.572",
            "/attr/Satellite Name": "FY4B", "/attr/Sensor Identification Code": "GHI", "/attr/Sensor Name": "GHI",
        }

        data = {}
        if self.filetype_info["file_type"] == "ghi_l1_0250m":
            data = self._get_250m_data("250")
        elif self.filetype_info["file_type"] == "ghi_l1_0500m":
            data = self._get_500m_data("500")
        elif self.filetype_info["file_type"] == "ghi_l1_2000m":
            data = self._get_2km_data("2000")
        elif self.filetype_info["file_type"] == "ghi_l1_2000m_geo":
            data = self._get_geo_data("2000")

        test_content = {}
        test_content.update(global_attrs)
        test_content.update(data)

        return test_content


def _create_filenames_from_resolutions(*resolutions):
    """Create filenames from the given resolutions."""
    if "GEO" in resolutions:
        return ["FY4B-_GHI---_N_REGX_1330E_L1-_GEO-_MULT_NOM_20220613145300_20220613145359_2000M_V0001.HDF"]
    pattern = ("FY4B-_GHI---_N_REGX_1330E_L1-_FDI-_MULT_NOM_20220613145300_20220613145359_"
               "{resolution:04d}M_V0001.HDF")
    return [pattern.format(resolution=resolution) for resolution in resolutions]


class Test_HDF_GHI_L1_cal:
    """Test VIRR L1B Reader."""

    yaml_file = "ghi_l1.yaml"

    def setup_method(self):
        """Wrap HDF5 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.core.fy4 import FY4Base
        from satpy.readers.ghi_l1 import HDF_GHI_L1
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.fy4 = mock.patch.object(FY4Base, "__bases__", (FakeHDF5FileHandler2,))
        self.p = mock.patch.object(HDF_GHI_L1.__class__, (self.fy4,))
        self.fake_handler = self.fy4.start()
        self.p.is_local = True

        self.expected = {
            "C01": np.array([[2.01, 2.02, 2.03, 2.04, 2.05], [2.06, 2.07, 2.08, 2.09, 2.1]]),
            "C02": np.array([[4.03, 4.06, 4.09, 4.12, 4.15], [4.18, 4.21, 4.24, 4.27, 4.3]]),
            "C03": np.array([[6.05, 6.1, 6.15, 6.2, 6.25], [6.3, 6.35, 6.4, 6.45, 6.5]]),
            "C04": np.array([[8.07, 8.14, 8.21, 8.28, 8.35], [8.42, 8.49, 8.56, 8.63, 8.7]]),
            "C05": np.array([[10.09, 10.18, 10.27, 10.36, 10.45], [10.54, 10.63, 10.72, 10.81, 10.9]]),
            "C06": np.array([[12.11, 12.22, 12.33, 12.44, 12.55], [12.66, 12.77, 12.88, 12.99, 13.1]]),
            "C07": np.array([[0.2, 0.3, 0.4, 0.5, 0.6], [0.7, 0.8, 0.9, 1., np.nan]]),
        }

    def teardown_method(self):
        """Stop wrapping the HDF5 file handler."""
        self.p.stop()

    def test_ghi_channels_are_loaded_with_right_resolution(self):
        """Test all channels are loaded with the right resolution."""
        reader = self._create_reader_for_resolutions(*RESOLUTION_LIST)

        available_datasets = reader.available_dataset_ids

        for resolution_to_test in RESOLUTION_LIST:
            self._check_keys_for_dsq(available_datasets, resolution_to_test)

    def test_ghi_all_bands_have_right_units(self):
        """Test all bands have the right units."""
        reader = self._create_reader_for_resolutions(*RESOLUTION_LIST)

        band_names = ALL_BAND_NAMES
        res = reader.load(band_names)
        assert len(res) == 7

        for band_name in band_names:
            assert res[band_name].shape == (2, 5)
            self._check_units(band_name, res)

    def test_ghi_orbital_parameters_are_correct(self):
        """Test orbital parameters are set correctly."""
        reader = self._create_reader_for_resolutions(*RESOLUTION_LIST)

        band_names = ALL_BAND_NAMES
        res = reader.load(band_names)

        # check whether the data type of orbital_parameters is float
        orbital_parameters = res[band_names[0]].attrs["orbital_parameters"]
        for attr in orbital_parameters:
            assert isinstance(orbital_parameters[attr], float)
        assert orbital_parameters["satellite_nominal_latitude"] == 0.
        assert orbital_parameters["satellite_nominal_longitude"] == 133.0
        assert orbital_parameters["satellite_nominal_altitude"] == 3.5786E7

    @staticmethod
    def _check_keys_for_dsq(available_datasets, resolution_to_test):
        from satpy.dataset.data_dict import get_key
        from satpy.tests.utils import make_dsq

        band_names = CHANNELS_BY_RESOLUTION[resolution_to_test]
        for band_name in band_names:
            ds_q = make_dsq(name=band_name, resolution=resolution_to_test)
            res = get_key(ds_q, available_datasets, num_results=0, best=False)
            if band_name < "C07":
                assert len(res) == 2
            else:
                assert len(res) == 3

    def test_ghi_counts_calibration(self):
        """Test loading data at counts calibration."""
        from satpy.tests.utils import make_dsq
        reader = self._create_reader_for_resolutions(*RESOLUTION_LIST)

        ds_ids = []
        band_names = CHANNELS_BY_RESOLUTION[2000]
        for band_name in band_names:
            ds_ids.append(make_dsq(name=band_name, calibration="counts"))
        res = reader.load(ds_ids)
        assert len(res) == 7

        for band_name in band_names:
            assert res[band_name].shape == (2, 5)
            assert res[band_name].attrs["calibration"] == "counts"
            assert res[band_name].dtype == np.uint16
            assert res[band_name].attrs["units"] == "1"

    def test_ghi_geo(self):
        """Test loading data for angles."""
        from satpy.tests.utils import make_dsq
        reader = self._create_reader_for_resolutions("GEO")
        band_name = "solar_azimuth_angle"
        ds_ids = [make_dsq(name=band_name)]
        res = reader.load(ds_ids)
        assert len(res) == 1

        assert res[band_name].shape == (2, 5)
        assert res[band_name].dtype == np.float32

    def _create_reader_for_resolutions(self, *resolutions):
        from satpy.readers.core.loading import load_reader
        filenames = _create_filenames_from_resolutions(*resolutions)
        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        assert len(filenames) == len(files)
        reader.create_filehandlers(files)
        # Make sure we have some files
        assert reader.file_handlers
        return reader

    @pytest.mark.parametrize("resolution_to_test", RESOLUTION_LIST)
    def test_ghi_for_one_resolution(self, resolution_to_test):
        """Test loading data when only one resolution is available."""
        reader = self._create_reader_for_resolutions(resolution_to_test)
        available_datasets = reader.available_dataset_ids
        band_names = CHANNELS_BY_RESOLUTION[resolution_to_test]
        self._assert_which_channels_are_loaded(available_datasets, band_names, resolution_to_test)
        res = reader.load(band_names)
        assert len(res) == len(band_names)
        self._check_calibration_and_units(band_names, res)
        for band_name in band_names:
            np.testing.assert_allclose(np.array(res[band_name].attrs["area"].area_extent),
                                       np.array(AREA_EXTENTS_BY_RESOLUTION[resolution_to_test]))

    def _check_calibration_and_units(self, band_names, result):
        for band_name in band_names:
            assert result[band_name].attrs["sensor"].islower()
            assert result[band_name].shape == (2, 5)
            np.testing.assert_allclose(result[band_name].values, self.expected[band_name], equal_nan=True)
            self._check_units(band_name, result)

    @staticmethod
    def _check_units(band_name, result):
        if band_name <= "C06":
            assert result[band_name].attrs["calibration"] == "reflectance"
        else:
            assert result[band_name].attrs["calibration"] == "brightness_temperature"
        if band_name <= "C06":
            assert result[band_name].attrs["units"] == "%"
        else:
            assert result[band_name].attrs["units"] == "K"

    @staticmethod
    def _assert_which_channels_are_loaded(available_datasets, band_names, resolution_to_test):
        from satpy.dataset.data_dict import get_key
        from satpy.tests.utils import make_dsq

        other_resolutions = RESOLUTION_LIST.copy()
        other_resolutions.remove(resolution_to_test)
        for band_name in band_names:
            for resolution in other_resolutions:
                ds_q = make_dsq(name=band_name, resolution=resolution)
                with pytest.raises(KeyError):
                    _ = get_key(ds_q, available_datasets, num_results=0, best=False)

            ds_q = make_dsq(name=band_name, resolution=resolution_to_test)
            res = get_key(ds_q, available_datasets, num_results=0, best=False)
            if band_name < "C07":
                assert len(res) == 2
            else:
                assert len(res) == 3



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for the dependency tree class and dependencies."""

import unittest
from unittest.mock import MagicMock

from satpy.node import CompositorNode


class FakeCompositor:
    """A fake compositor."""

    def __init__(self, id):  # noqa: A002
        """Set up the fake compositor."""
        self.id = id


class TestCompositorNodeCopy(unittest.TestCase):
    """Test case for copying a node."""

    def setUp(self):
        """Set up the test case."""
        self.node = CompositorNode(MagicMock())
        self.node.add_required_nodes([MagicMock(), MagicMock()])
        self.node.add_optional_nodes([MagicMock()])

        self.node_copy = self.node.copy()

    def test_node_data_is_copied(self):
        """Test that the data of the node is copied."""
        assert self.node_copy.data is not self.node.data

    def test_node_data_required_nodes_are_copies(self):
        """Test that the required nodes of the node data are copied."""
        for req1, req2 in zip(self.node.required_nodes, self.node_copy.required_nodes):
            assert req1 is not req2

    def test_node_data_optional_nodes_are_copies(self):
        """Test that the optional nodes of the node data are copied."""
        for req1, req2 in zip(self.node.optional_nodes, self.node_copy.optional_nodes):
            assert req1 is not req2


class TestCompositorNode(unittest.TestCase):
    """Test case for the compositor node object."""

    def setUp(self):
        """Set up the test case."""
        self.name = "hej"
        self.fake = FakeCompositor(self.name)
        self.c_node = CompositorNode(self.fake)

    def test_compositor_node_init(self):
        """Test compositor node initialization."""
        assert self.c_node.name == self.name
        assert self.fake in self.c_node.data

    def test_add_required_nodes(self):
        """Test adding required nodes."""
        self.c_node.add_required_nodes([1, 2, 3])
        assert self.c_node.required_nodes == [1, 2, 3]

    def test_add_required_nodes_twice(self):
        """Test adding required nodes twice."""
        self.c_node.add_required_nodes([1, 2])
        self.c_node.add_required_nodes([3])
        assert self.c_node.required_nodes == [1, 2, 3]

    def test_add_optional_nodes(self):
        """Test adding optional nodes."""
        self.c_node.add_optional_nodes([1, 2, 3])
        assert self.c_node.optional_nodes == [1, 2, 3]

    def test_add_optional_nodes_twice(self):
        """Test adding optional nodes twice."""
        self.c_node.add_optional_nodes([1, 2])
        self.c_node.add_optional_nodes([3])
        assert self.c_node.optional_nodes == [1, 2, 3]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2015-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Behaviour steps for loading."""

import os
from urllib.request import urlopen

from behave import given, then, use_step_matcher, when

use_step_matcher("re")


@given(u"data is available")
def step_impl_data_available(context):
    """Make data available."""
    if not os.path.exists("/tmp/SVM02_npp_d20150311_t1122204_e1123446_b17451_c20150311113206961730_cspp_dev.h5"):
        response = urlopen("https://zenodo.org/record/16355/files/"
                           "SVM02_npp_d20150311_t1122204_e1123446_b17451_c20150311113206961730_cspp_dev.h5")
        with open("/tmp/SVM02_npp_d20150311_t1122204_e1123446_b17451_c20150311113206961730_cspp_dev.h5",
                  mode="w") as fp:
            fp.write(response.read())
    if not os.path.exists("/tmp/GMTCO_npp_d20150311_t1122204_e1123446_b17451_c20150311113205873710_cspp_dev.h5"):
        response = urlopen("https://zenodo.org/record/16355/files/"
                           "GMTCO_npp_d20150311_t1122204_e1123446_b17451_c20150311113205873710_cspp_dev.h5")
        with open("/tmp/GMTCO_npp_d20150311_t1122204_e1123446_b17451_c20150311113205873710_cspp_dev.h5",
                  mode="w") as fp:
            fp.write(response.read())


@when(u"user loads the data without providing a config file")
def step_impl_user_loads_no_config(context):
    """Load the data without a config."""
    import datetime as dt

    from satpy import Scene, find_files_and_readers
    os.chdir("/tmp/")
    readers_files = find_files_and_readers(sensor="viirs",
                                           start_time=dt.datetime(2015, 3, 11, 11, 20),
                                           end_time=dt.datetime(2015, 3, 11, 11, 26))
    scn = Scene(filenames=readers_files)
    scn.load(["M02"])
    context.scene = scn


@then(u"the data is available in a scene object")
def step_impl_data_available_in_scene(context):
    """Check that the data is available in the scene."""
    assert context.scene["M02"] is not None
    assert context.scene.get("M01") is None


@when(u"some items are not available")
def step_impl_items_not_available(context):
    """Load some data."""
    context.scene.load(["M01"])


@when(u"user wants to know what data is available")
def step_impl_user_checks_availability(context):
    """Check availability."""
    import datetime as dt

    from satpy import Scene, find_files_and_readers
    os.chdir("/tmp/")
    reader_files = find_files_and_readers(sensor="viirs",
                                          start_time=dt.datetime(2015, 3, 11, 11, 20),
                                          end_time=dt.datetime(2015, 3, 11, 11, 26))
    scn = Scene(filenames=reader_files)
    context.available_dataset_ids = scn.available_dataset_ids()


@then(u"available datasets are returned")
def step_impl_available_datasets_are_returned(context):
    """Check that datasets are returned."""
    assert (len(context.available_dataset_ids) >= 5)


@given("datasets with the same name")
def step_impl_datasets_with_same_name(context):
    """Datasets with the same name but different other ID parameters."""
    from xarray import DataArray

    from satpy import Scene
    from satpy.tests.utils import make_dataid
    scn = Scene()
    scn[make_dataid(name="ds1", calibration="radiance")] = DataArray([[1, 2], [3, 4]])
    scn[make_dataid(name="ds1", resolution=500, calibration="reflectance")] = DataArray([[5, 6], [7, 8]])
    scn[make_dataid(name="ds1", resolution=250, calibration="reflectance")] = DataArray([[5, 6], [7, 8]])
    scn[make_dataid(name="ds1", resolution=1000, calibration="reflectance")] = DataArray([[5, 6], [7, 8]])
    scn[make_dataid(name="ds1", resolution=500, calibration="radiance", modifiers=("mod1",))] = \
        DataArray([[5, 6], [7, 8]])
    ds_id = make_dataid(name="ds1", resolution=1000, calibration="radiance", modifiers=("mod1", "mod2"))
    scn[ds_id] = DataArray([[5, 6], [7, 8]])
    context.scene = scn


@when("a dataset is retrieved by name")
def step_impl_dataset_retrieved_by_name(context):
    """Use the Scene's getitem method to get a dataset."""
    context.returned_dataset = context.scene["ds1"]


@then("the least modified version of the dataset is returned")
def step_impl_least_modified_dataset_returned(context):
    """Check that the dataset should be one of the least modified datasets."""
    assert len(context.returned_dataset.attrs["modifiers"]) == 0



# Copyright (c) 2022 Pytroll Developers

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The mws_l1b_nc reader tests.

This module tests the reading of the MWS l1b netCDF format data as per version v4B issued 22 November 2021.

"""

import datetime as dt
import logging
from unittest.mock import patch

import numpy as np
import pytest
import xarray as xr
from netCDF4 import Dataset

from satpy.readers.mws_l1b import MWSL1BFile, get_channel_index_from_name

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path
# - caplog

N_CHANNELS = 24
N_CHANNELS_OS = 2
N_SCANS = 2637
N_FOVS = 95
N_FOVS_CAL = 5
N_PRTS = 6


@pytest.fixture
def reader(fake_file):
    """Return reader of mws level-1b data."""
    return MWSL1BFile(
        filename=fake_file,
        filename_info={
            "start_time": (
                dt.datetime.fromisoformat("2000-01-01T01:00:00")
            ),
            "end_time": (
                dt.datetime.fromisoformat("2000-01-01T02:00:00")
            ),
            "creation_time": (
                dt.datetime.fromisoformat("2000-01-01T03:00:00")
            ),
        },
        filetype_info={
            "longitude": "data/navigation_data/mws_lon",
            "latitude": "data/navigation_data/mws_lat",
            "solar_azimuth": "data/navigation/mws_solar_azimuth_angle",
            "solar_zenith": "data/navigation/mws_solar_zenith_angle",
            "satellite_azimuth": "data/navigation/mws_satellite_azimuth_angle",
            "satellite_zenith": "data/navigation/mws_satellite_zenith_angle",
        }
    )


@pytest.fixture
def fake_file(tmp_path):
    """Return file path to level-1b file."""
    file_path = tmp_path / "test_file_mws_l1b.nc"
    writer = MWSL1BFakeFileWriter(file_path)
    writer.write()
    return file_path


class MWSL1BFakeFileWriter:
    """Writer class of fake mws level-1b data."""

    def __init__(self, file_path):
        """Init."""
        self.file_path = file_path

    def write(self):
        """Write fake data to file."""
        with Dataset(self.file_path, "w") as dataset:
            self._write_attributes(dataset)
            self._write_status_group(dataset)
            self._write_quality_group(dataset)
            data_group = dataset.createGroup("data")
            self._create_scan_dimensions(data_group)
            self._write_navigation_data_group(data_group)
            self._write_calibration_data_group(data_group)
            self._write_measurement_data_group(data_group)

    @staticmethod
    def _write_attributes(dataset):
        """Write attributes."""
        dataset.sensing_start_time_utc = "2000-01-02 03:04:05.000"
        dataset.sensing_end_time_utc = "2000-01-02 04:05:06.000"
        dataset.instrument = "MWS"
        dataset.spacecraft = "SGA1"

    @staticmethod
    def _write_status_group(dataset):
        """Write the status group."""
        group = dataset.createGroup("/status/satellite")
        subsat_latitude_start = group.createVariable(
            "subsat_latitude_start", "f4"
        )
        subsat_latitude_start[:] = 52.19

        subsat_longitude_start = group.createVariable(
            "subsat_longitude_start", "f4"
        )
        subsat_longitude_start[:] = 23.26

        subsat_latitude_end = group.createVariable(
            "subsat_latitude_end", "f4"
        )
        subsat_latitude_end[:] = 60.00

        subsat_longitude_end = group.createVariable(
            "subsat_longitude_end", "f4"
        )
        subsat_longitude_end[:] = 2.47

    @staticmethod
    def _write_quality_group(dataset):
        """Write the quality group."""
        group = dataset.createGroup("quality")
        group.overall_quality_flag = 0
        duration_of_product = group.createVariable(
            "duration_of_product", "f4"
        )
        duration_of_product[:] = 5944.

    @staticmethod
    def _write_navigation_data_group(dataset):
        """Write the navigation data group."""
        group = dataset.createGroup("navigation")
        dimensions = ("n_scans", "n_fovs")
        shape = (N_SCANS, N_FOVS)
        longitude = group.createVariable(
            "mws_lon",
            np.int32,
            dimensions=dimensions,
        )
        longitude.scale_factor = 1.0E-4
        longitude.add_offset = 0.0
        longitude.missing_value = np.array((-2147483648), np.int32)
        longitude[:] = 35.7535 * np.ones(shape)

        latitude = group.createVariable(
            "mws_lat",
            np.float32,
            dimensions=dimensions,
        )
        latitude[:] = 2. * np.ones(shape)

        azimuth = group.createVariable(
            "mws_solar_azimuth_angle",
            np.float32,
            dimensions=dimensions,
        )
        azimuth[:] = 179. * np.ones(shape)

    @staticmethod
    def _create_scan_dimensions(dataset):
        """Create the scan/fovs dimensions."""
        dataset.createDimension("n_channels", N_CHANNELS)
        dataset.createDimension("n_channels_os", N_CHANNELS_OS)
        dataset.createDimension("n_scans", N_SCANS)
        dataset.createDimension("n_fovs", N_FOVS)
        dataset.createDimension("n_prts", N_PRTS)
        dataset.createDimension("n_fovs_cal", N_FOVS_CAL)

    @staticmethod
    def _write_calibration_data_group(dataset):
        """Write the calibration data group."""
        group = dataset.createGroup("calibration")
        toa_bt = group.createVariable(
            "mws_toa_brightness_temperature", np.float32, dimensions=("n_scans", "n_fovs", "n_channels",)
        )
        toa_bt.scale_factor = 1.0  # 1.0E-8
        toa_bt.add_offset = 0.0
        toa_bt.missing_value = -2147483648
        toa_bt[:] = 240.0 * np.ones((N_SCANS, N_FOVS, N_CHANNELS))

    @staticmethod
    def _write_measurement_data_group(dataset):
        """Write the measurement data group."""
        group = dataset.createGroup("measurement")
        counts = group.createVariable(
            "mws_earth_view_counts", np.int32, dimensions=("n_scans", "n_fovs", "n_channels",)
        )
        counts[:] = 24100 * np.ones((N_SCANS, N_FOVS, N_CHANNELS), dtype=np.int32)


class TestMwsL1bNCFileHandler:
    """Test the MWSL1BFile reader."""

    def test_start_time(self, reader):
        """Test acquiring the start time."""
        assert reader.start_time == dt.datetime(2000, 1, 2, 3, 4, 5)

    def test_end_time(self, reader):
        """Test acquiring the end time."""
        assert reader.end_time == dt.datetime(2000, 1, 2, 4, 5, 6)

    def test_sensor(self, reader):
        """Test sensor."""
        assert reader.sensor == "MWS"

    def test_platform_name(self, reader):
        """Test getting the platform name."""
        assert reader.platform_name == "Metop-SG-A1"

    def test_sub_satellite_longitude_start(self, reader):
        """Test getting the longitude of sub-satellite point at start of the product."""
        np.testing.assert_allclose(reader.sub_satellite_longitude_start, 23.26)

    def test_sub_satellite_latitude_start(self, reader):
        """Test getting the latitude of sub-satellite point at start of the product."""
        np.testing.assert_allclose(reader.sub_satellite_latitude_start, 52.19)

    def test_sub_satellite_longitude_end(self, reader):
        """Test getting the longitude of sub-satellite point at end of the product."""
        np.testing.assert_allclose(reader.sub_satellite_longitude_end, 2.47)

    def test_sub_satellite_latitude_end(self, reader):
        """Test getting the latitude of sub-satellite point at end of the product."""
        np.testing.assert_allclose(reader.sub_satellite_latitude_end, 60.0)

    def test_get_dataset_get_channeldata_counts(self, reader):
        """Test getting channel data."""
        dataset_id = {"name": "1", "units": None,
                      "calibration": "counts"}
        dataset_info = {"file_key": "data/measurement/mws_earth_view_counts"}

        dataset = reader.get_dataset(dataset_id, dataset_info)
        expected_bt = np.array([[24100, 24100],
                                [24100, 24100]], dtype=np.int32)
        count = dataset[10:12, 12:14].data.compute()
        np.testing.assert_allclose(count, expected_bt)

    def test_get_dataset_get_channeldata_bts(self, reader):
        """Test getting channel data."""
        dataset_id = {"name": "1", "units": "K",
                      "calibration": "brightness_temperature"}
        dataset_info = {"file_key": "data/calibration/mws_toa_brightness_temperature"}

        dataset = reader.get_dataset(dataset_id, dataset_info)

        expected_bt = np.array([[240., 240., 240., 240., 240.],
                                [240., 240., 240., 240., 240.],
                                [240., 240., 240., 240., 240.],
                                [240., 240., 240., 240., 240.],
                                [240., 240., 240., 240., 240.]], dtype=np.float32)

        toa_bt = dataset[0:5, 0:5].data.compute()
        np.testing.assert_allclose(toa_bt, expected_bt)

    def test_get_dataset_return_none_if_data_not_exist(self, reader):
        """Test get dataset return none if data does not exist."""
        dataset_id = {"name": "unknown"}
        dataset_info = {"file_key": "non/existing/data"}
        dataset = reader.get_dataset(dataset_id, dataset_info)
        assert dataset is None

    def test_get_navigation_longitudes(self, caplog, fake_file, reader):
        """Test get the longitudes."""
        dataset_id = {"name": "mws_lon"}
        dataset_info = {"file_key": "data/navigation_data/mws_lon"}

        dataset = reader.get_dataset(dataset_id, dataset_info)

        expected_lons = np.array([[35.753498, 35.753498, 35.753498, 35.753498, 35.753498],
                                  [35.753498, 35.753498, 35.753498, 35.753498, 35.753498],
                                  [35.753498, 35.753498, 35.753498, 35.753498, 35.753498],
                                  [35.753498, 35.753498, 35.753498, 35.753498, 35.753498],
                                  [35.753498, 35.753498, 35.753498, 35.753498, 35.753498]], dtype=np.float32)

        longitudes = dataset[0:5, 0:5].data.compute()
        np.testing.assert_allclose(longitudes, expected_lons)

    def test_get_dataset_logs_debug_message(self, caplog, fake_file, reader):
        """Test get dataset return none if data does not exist."""
        dataset_id = {"name": "mws_lon"}
        dataset_info = {"file_key": "data/navigation_data/mws_lon"}

        with caplog.at_level(logging.DEBUG):
            _ = reader.get_dataset(dataset_id, dataset_info)

        log_output = "Reading mws_lon from {filename}".format(filename=str(fake_file))
        assert log_output in caplog.text

    def test_get_dataset_aux_data_not_supported(self, reader):
        """Test get auxillary dataset not supported."""
        dataset_id = {"name": "scantime_utc"}
        dataset_info = {"file_key": "non/existing"}

        with pytest.raises(NotImplementedError) as exec_info:
            _ = reader.get_dataset(dataset_id, dataset_info)

        assert str(exec_info.value) == "Dataset 'scantime_utc' not supported!"

    def test_get_dataset_aux_data_expected_data_missing(self, caplog, reader):
        """Test get auxillary dataset which is not present but supposed to be in file."""
        dataset_id = {"name": "surface_type"}
        dataset_info = {"file_key": "non/existing"}

        with caplog.at_level(logging.ERROR):
            with pytest.raises(KeyError) as exec_info:
                _ = reader.get_dataset(dataset_id, dataset_info)

        assert str(exec_info.value) == "'data/navigation/mws_surface_type'"

        log_output = ("Could not find key data/navigation/mws_surface_type in NetCDF file," +
                      " no valid Dataset created")
        assert log_output in caplog.text

    @pytest.mark.parametrize("dims", [
        ("n_scans", "n_fovs"),
        ("x", "y"),
    ])
    def test_standardize_dims(self, reader, dims):
        """Test standardize dims."""
        variable = xr.DataArray(
            np.arange(6).reshape(2, 3),
            dims=dims,
        )
        standardized = reader._standardize_dims(variable)
        assert standardized.dims == ("y", "x")

    @staticmethod
    def test_drop_coords(reader):
        """Test drop coordinates."""
        coords = "dummy"
        data = xr.DataArray(
            np.ones(10),
            dims=("y"),
            coords={coords: 0},
        )
        assert coords in data.coords
        data = reader._drop_coords(data)
        assert coords not in data.coords

    def test_get_global_attributes(self, reader):
        """Test get global attributes."""
        attributes = reader._get_global_attributes()
        assert attributes == {
            "filename": reader.filename,
            "start_time": dt.datetime(2000, 1, 2, 3, 4, 5),
            "end_time": dt.datetime(2000, 1, 2, 4, 5, 6),
            "spacecraft_name": "Metop-SG-A1",
            "sensor": "MWS",
            "filename_start_time": dt.datetime(2000, 1, 1, 1, 0),
            "filename_end_time": dt.datetime(2000, 1, 1, 2, 0),
            "platform_name": "Metop-SG-A1",
            "quality_group": {
                "duration_of_product": np.array(5944., dtype=np.float32),
                "overall_quality_flag": 0,
            }
        }

    @patch(
        "satpy.readers.mws_l1b.MWSL1BFile._get_global_attributes",
        return_value={"mocked_global_attributes": True},
    )
    def test_manage_attributes(self, mock, reader):
        """Test manage attributes."""
        variable = xr.DataArray(
            np.ones(N_SCANS),
            attrs={"season": "summer"},
        )
        dataset_info = {"name": "1", "units": "K"}
        variable = reader._manage_attributes(variable, dataset_info)
        assert variable.attrs == {
            "season": "summer",
            "units": "K",
            "name": "1",
            "mocked_global_attributes": True,
        }


@pytest.mark.parametrize(("name", "index"), [("1", 0), ("2", 1), ("24", 23)])
def test_get_channel_index_from_name(name, index):
    """Test getting the MWS channel index from the channel name."""
    ch_idx = get_channel_index_from_name(name)
    assert ch_idx == index


def test_get_channel_index_from_name_throw_exception():
    """Test that an excpetion is thrown when getting the MWS channel index from an unsupported name."""
    with pytest.raises(AttributeError, match="Channel name 'channel 1' not supported"):
        _ = get_channel_index_from_name("channel 1")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Benchmark utilities."""

import os


def get_filenames(subdir):
    """Get the data filenames manually."""
    import glob
    base_dir = os.environ.get("SATPY_DEMO_DATA_DIR", ".")
    return glob.glob(os.path.join(base_dir, subdir, "*"))


class GeoBenchmarks:
    """Class for geo benchmarks."""

    def create_scene(self, filenames=None):
        """Create a scene."""
        from satpy import Scene
        scn = Scene(filenames=filenames or self.data_files, reader=self.reader)
        return scn

    def load_no_padding(self, composite, filenames=None):
        """Load one composite or channel."""
        scn = self.create_scene(filenames=filenames)
        scn.load([composite], pad_data=False)
        return scn

    def load_and_native_resample(self, composite):
        """Load and native resample a composite or channel."""
        return self.load_and_resample(composite, "native")

    def load_and_resample(self, composite, resampler, area=None, filenames=None):
        """Load and resample a composite or channel with resampler and area."""
        scn = self.load_no_padding(composite, filenames=filenames)
        ls = scn.resample(area, resampler=resampler)
        ls._readers = scn._readers  # workaround for GH#1861
        return ls

    def compute_composite(self, composite, resampler="native",
                          area=None, filenames=None):
        """Compute a true color image."""
        lscn = self.load_and_resample(
                composite, resampler, area, filenames)
        lscn[composite].compute()

    def save_composite_as_geotiff(self, composite, resampler="native",
                                  area=None, filenames=None):
        """Save a composite to disk as geotiff."""
        lscn = self.load_and_resample(composite, resampler, area, filenames)
        lscn.save_dataset(composite, filename="test.tif", tiled=True)

    def compute_channel(self, channel, filenames=None):
        """Load and compute one channel."""
        scn = self.load_no_padding(channel, filenames=filenames)
        scn[channel].compute()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2022, 2023 Satpy Developers

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Common utilities for reading VIIRS and ATMS SDR data."""

import datetime as dt
import logging

import dask.array as da
import numpy as np
import xarray as xr

from satpy.readers.core.hdf5 import HDF5FileHandler

NO_DATE = dt.datetime(1958, 1, 1)
EPSILON_TIME = dt.timedelta(days=2)
LOG = logging.getLogger(__name__)


VIIRS_DATASET_KEYS = {"GDNBO": "VIIRS-DNB-GEO",
                      "SVDNB": "VIIRS-DNB-SDR",
                      "GITCO": "VIIRS-IMG-GEO-TC",
                      "GIMGO": "VIIRS-IMG-GEO",
                      "SVI01": "VIIRS-I1-SDR",
                      "SVI02": "VIIRS-I2-SDR",
                      "SVI03": "VIIRS-I3-SDR",
                      "SVI04": "VIIRS-I4-SDR",
                      "SVI05": "VIIRS-I5-SDR",
                      "GMTCO": "VIIRS-MOD-GEO-TC",
                      "GMODO": "VIIRS-MOD-GEO",
                      "SVM01": "VIIRS-M1-SDR",
                      "SVM02": "VIIRS-M2-SDR",
                      "SVM03": "VIIRS-M3-SDR",
                      "SVM04": "VIIRS-M4-SDR",
                      "SVM05": "VIIRS-M5-SDR",
                      "SVM06": "VIIRS-M6-SDR",
                      "SVM07": "VIIRS-M7-SDR",
                      "SVM08": "VIIRS-M8-SDR",
                      "SVM09": "VIIRS-M9-SDR",
                      "SVM10": "VIIRS-M10-SDR",
                      "SVM11": "VIIRS-M11-SDR",
                      "SVM12": "VIIRS-M12-SDR",
                      "SVM13": "VIIRS-M13-SDR",
                      "SVM14": "VIIRS-M14-SDR",
                      "SVM15": "VIIRS-M15-SDR",
                      "SVM16": "VIIRS-M16-SDR",
                      "IVCDB": "VIIRS-DualGain-Cal-IP"}
ATMS_DATASET_KEYS = {"SATMS": "ATMS-SDR",
                     "GATMO": "ATMS-SDR-GEO",
                     "TATMS": "ATMS-TDR"}

DATASET_KEYS = {}
DATASET_KEYS.update(VIIRS_DATASET_KEYS)
DATASET_KEYS.update(ATMS_DATASET_KEYS)


def _get_scale_factors_for_units(factors, file_units, output_units):
    if file_units == "W cm-2 sr-1" and output_units == "W m-2 sr-1":
        LOG.debug("Adjusting scaling factors to convert '%s' to '%s'",
                  file_units, output_units)
        factors = factors * 10000.
    elif file_units == "1" and output_units == "%":
        LOG.debug("Adjusting scaling factors to convert '%s' to '%s'",
                  file_units, output_units)
        factors = factors * 100.
    else:
        raise ValueError("Don't know how to convert '{}' to '{}'".format(
            file_units, output_units))
    return factors


def _get_file_units(dataset_id, ds_info):
    """Get file units from metadata."""
    file_units = ds_info.get("file_units")
    if file_units is None:
        LOG.debug("Unknown units for file key '%s'", dataset_id)
    return file_units


class JPSS_SDR_FileHandler(HDF5FileHandler):
    """Base class for reading JPSS VIIRS & ATMS SDR HDF5 Files."""

    def __init__(self, filename, filename_info, filetype_info, **kwargs):
        """Initialize file handler."""
        super().__init__(filename, filename_info, filetype_info, **kwargs)

    def _parse_datetime(self, datestr, timestr):
        if not isinstance(datestr, str):
            datestr = str(datestr.data.compute().astype(str))
        if not isinstance(timestr, str):
            timestr = str(timestr.data.compute().astype(str))
        datetime_str = datestr + timestr

        time_val = dt.datetime.strptime(datetime_str, "%Y%m%d%H%M%S.%fZ")
        if abs(time_val - NO_DATE) < EPSILON_TIME:
            # catch rare case when SDR files have incorrect date
            raise ValueError("Datetime invalid {}".format(time_val))
        return time_val

    @property
    def start_time(self):
        """Get start time."""
        date_var_path = self._get_aggr_path("start_date", "AggregateBeginningDate")
        time_var_path = self._get_aggr_path("start_time", "AggregateBeginningTime")
        return self._parse_datetime(self[date_var_path], self[time_var_path])

    @property
    def end_time(self):
        """Get end time."""
        date_var_path = self._get_aggr_path("end_date", "AggregateEndingDate")
        time_var_path = self._get_aggr_path("end_time", "AggregateEndingTime")
        return self._parse_datetime(self[date_var_path], self[time_var_path])

    @property
    def start_orbit_number(self):
        """Get start orbit number."""
        start_orbit_path = self._get_aggr_path("start_orbit", "AggregateBeginningOrbitNumber")
        return int(self[start_orbit_path])

    @property
    def end_orbit_number(self):
        """Get end orbit number."""
        end_orbit_path = self._get_aggr_path("end_orbit", "AggregateEndingOrbitNumber")
        return int(self[end_orbit_path])

    def _get_aggr_path(self, fileinfo_key, aggr_default):
        dataset_group = DATASET_KEYS[self.datasets[0]]
        default = "Data_Products/{dataset_group}/{dataset_group}_Aggr/attr/" + aggr_default
        return self.filetype_info.get(fileinfo_key, default).format(dataset_group=dataset_group)

    @property
    def platform_name(self):
        """Get platform name."""
        default = "/attr/Platform_Short_Name"
        platform_path = self.filetype_info.get(
            "platform_name", default).format(**self.filetype_info)
        platform_dict = {"NPP": "Suomi-NPP",
                         "JPSS-1": "NOAA-20",
                         "J01": "NOAA-20",
                         "JPSS-2": "NOAA-21",
                         "J02": "NOAA-21"}
        return platform_dict.get(self[platform_path], self[platform_path])

    @property
    def sensor_name(self):
        """Get sensor name."""
        dataset_group = DATASET_KEYS[self.datasets[0]]
        default = "Data_Products/{dataset_group}/attr/Instrument_Short_Name"
        sensor_path = self.filetype_info.get(
            "sensor_name", default).format(dataset_group=dataset_group)
        return self[sensor_path].lower()

    def scale_swath_data(self, data, scaling_factors, dataset_group):
        """Scale swath data using scaling factors and offsets.

        Multi-granule (a.k.a. aggregated) files will have more than the usual two values.
        """
        rows_per_gran = self._get_rows_per_granule(dataset_group)
        factors = self._mask_and_reshape_factors(scaling_factors)
        data = self._map_and_apply_factors(data, factors, rows_per_gran)
        return data

    def scale_data_to_specified_unit(self, data, dataset_id, ds_info):
        """Get sscale and offset factors and convert/scale data to given physical unit."""
        var_path = self._generate_file_key(dataset_id, ds_info)
        dataset_group = ds_info["dataset_group"]
        file_units = _get_file_units(dataset_id, ds_info)
        output_units = ds_info.get("units", file_units)

        factor_var_path = ds_info.get("factors_key", var_path + "Factors")

        factors = self.get(factor_var_path)
        factors = self._adjust_scaling_factors(factors, file_units, output_units)

        if factors is not None:
            return self.scale_swath_data(data, factors, dataset_group)

        LOG.debug("No scaling factors found for %s", dataset_id)
        return data

    @staticmethod
    def _mask_and_reshape_factors(factors):
        factors = factors.where(factors > -999, np.float32(np.nan))
        return factors.data.reshape((-1, 2)).rechunk((1, 2))  # make it so map_blocks happens per factor

    @staticmethod
    def _map_and_apply_factors(data, factors, rows_per_gran):
        # The user may have requested a different chunking scheme, but we need
        # per granule chunking right now so factor chunks map 1:1 to data chunks
        old_chunks = data.chunks
        dask_data = data.data.rechunk((tuple(rows_per_gran), data.data.chunks[1]))
        dask_data = da.map_blocks(_apply_factors, dask_data, factors,
                                  chunks=dask_data.chunks, dtype=data.dtype,
                                  meta=np.array([[]], dtype=data.dtype))
        data = xr.DataArray(dask_data.rechunk(old_chunks),
                            dims=data.dims, coords=data.coords,
                            attrs=data.attrs)
        return data

    @staticmethod
    def _scale_factors_for_units(factors, file_units, output_units):
        return _get_scale_factors_for_units(factors, file_units, output_units)

    @staticmethod
    def _get_valid_scaling_factors(factors):
        if factors is None:
            factors = np.array([1, 0], dtype=np.float32)
            factors = xr.DataArray(da.from_array(factors, chunks=1))
        else:
            factors = factors.where(factors != -999., np.float32(np.nan))
        return factors

    def _adjust_scaling_factors(self, factors, file_units, output_units):
        """Adjust scaling factors ."""
        if file_units == output_units:
            LOG.debug("File units and output units are the same (%s)", file_units)
            return factors
        factors = self._get_valid_scaling_factors(factors)
        return self._scale_factors_for_units(factors, file_units, output_units)

    @staticmethod
    def expand_single_values(var, scans):
        """Expand single valued variable to full scan lengths."""
        if scans.size == 1:
            return var
        else:
            expanded = np.repeat(var, scans)
            expanded.attrs = var.attrs
            expanded.rename({expanded.dims[0]: "y"})
            return expanded

    def _scan_size(self, dataset_group_name):
        """Get how many rows of data constitute one scanline."""
        if "ATM" in dataset_group_name:
            scan_size = 1
        elif "I" in dataset_group_name:
            scan_size = 32
        else:
            scan_size = 16
        return scan_size

    def _generate_file_key(self, ds_id, ds_info, factors=False):
        var_path = ds_info.get("file_key", "All_Data/{dataset_group}_All/{calibration}")
        calibration = {
            "radiance": "Radiance",
            "reflectance": "Reflectance",
            "brightness_temperature": "BrightnessTemperature",
        }.get(ds_id.get("calibration"))
        var_path = var_path.format(calibration=calibration, dataset_group=DATASET_KEYS[ds_info["dataset_group"]])
        if ds_id["name"] in ["dnb_longitude", "dnb_latitude"]:
            if self.use_tc is True:
                return var_path + "_TC"
            if self.use_tc is None and var_path + "_TC" in self.file_content:
                return var_path + "_TC"
        return var_path

    def _update_data_attributes(self, data, dataset_id, ds_info):
        file_units = _get_file_units(dataset_id, ds_info)
        output_units = ds_info.get("units", file_units)
        i = getattr(data, "attrs", {})
        i.update(ds_info)
        i.update({
            "platform_name": self.platform_name,
            "sensor": self.sensor_name,
            "start_orbit": self.start_orbit_number,
            "end_orbit": self.end_orbit_number,
            "units": output_units,
            "rows_per_scan": self._scan_size(ds_info["dataset_group"]),
        })
        i.update(dataset_id.to_dict())
        data.attrs.update(i)
        return data

    def _get_variable(self, var_path, **kwargs):
        return self[var_path]

    def concatenate_dataset(self, dataset_group, var_path, **kwargs):
        """Concatenate dataset."""
        scan_size = self._scan_size(dataset_group)
        scans = self._get_scans_per_granule(dataset_group)
        start_scan = 0
        data_chunks = []
        scans = xr.DataArray(scans)

        variable = self._get_variable(var_path, **kwargs)
        # check if these are single per-granule value
        if variable.size != scans.size:
            for gscans in scans.values:
                data_chunks.append(variable.isel(y=slice(start_scan,
                                                         start_scan + gscans * scan_size)))
                start_scan += gscans * scan_size
            return xr.concat(data_chunks, "y")
        else:
            # This is not tested - Not sure this code is ever going to be used? A. Dybbroe
            # Mon Jan  2 13:31:21 2023
            return self.expand_single_values(variable, scans)

    def _get_rows_per_granule(self, dataset_group):
        scan_size = self._scan_size(dataset_group)
        scans_per_gran = self._get_scans_per_granule(dataset_group)
        return [scan_size * gran_scans for gran_scans in scans_per_gran]

    def _get_scans_per_granule(self, dataset_group):
        number_of_granules_path = "Data_Products/{dataset_group}/{dataset_group}_Aggr/attr/AggregateNumberGranules"
        nb_granules_path = number_of_granules_path.format(dataset_group=DATASET_KEYS[dataset_group])
        scans = []
        for granule in range(self[nb_granules_path]):
            scans_path = "Data_Products/{dataset_group}/{dataset_group}_Gran_{granule}/attr/N_Number_Of_Scans"
            scans_path = scans_path.format(dataset_group=DATASET_KEYS[dataset_group], granule=granule)
            scans.append(self[scans_path])
        return scans

    def mask_fill_values(self, data, ds_info):
        """Mask fill values."""
        is_floating = np.issubdtype(data.dtype, np.floating)

        if is_floating:
            # If the data is a float then we mask everything <= -999.0
            fill_max = np.float32(ds_info.pop("fill_max_float", -999.0))
            return data.where(data > fill_max, np.float32(np.nan))
        else:
            # If the data is an integer then we mask everything >= fill_min_int
            fill_min = int(ds_info.pop("fill_min_int", 65528))
            return data.where(data < fill_min, np.float32(np.nan))

    def available_datasets(self, configured_datasets=None):
        """Generate dataset info and their availablity.

        See
        :meth:`satpy.readers.core.file_handlers.BaseFileHandler.available_datasets`
        for details.

        """
        for is_avail, ds_info in (configured_datasets or []):
            if is_avail is not None:
                yield is_avail, ds_info
                continue
            dataset_group = [ds_group for ds_group in ds_info["dataset_groups"] if ds_group in self.datasets]
            if dataset_group:
                yield True, ds_info
            elif is_avail is None:
                yield is_avail, ds_info


def _apply_factors(data, factor_set):
    return data * factor_set[0, 0] + factor_set[0, 1]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Test roundripping the cf writer and reader."""

import os

import numpy as np
import pytest

from satpy import Scene
from satpy.tests.reader_tests.test_viirs_compact import fake_dnb, fake_dnb_file  # noqa

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path


# ignore that the "coordinates" added by the reader aren't found in the existing list of variables
# the correct lon/lat coordinates are listed in the final netcdf file
@pytest.mark.filterwarnings("ignore:Coordinate .*itude_dnb.* referenced by dataarray DNB:UserWarning")
def test_cf_roundtrip(fake_dnb_file, tmp_path):  # noqa
    """Test the cf writing reading cycle."""
    dnb_filename = os.fspath(fake_dnb_file)
    write_scn = Scene(filenames=[dnb_filename], reader="viirs_compact")
    write_scn.load(["DNB"])

    satpy_cf_file = os.fspath(tmp_path / "npp-viirs-20191025061125-20191025061247.nc")
    write_scn.save_datasets(writer="cf", filename=satpy_cf_file)
    read_scn = Scene(filenames=[satpy_cf_file], reader="satpy_cf_nc")
    read_scn.load(["DNB"])

    write_array = write_scn["DNB"]
    read_array = read_scn["DNB"]

    np.testing.assert_allclose(write_array.values, read_array.values)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Reader for SEADAS L2 products.

This reader currently only supports MODIS and VIIRS Chlorophyll A from SEADAS.

The reader includes an additional keyword argument ``apply_quality_flags``
which can be used to mask out low-quality pixels based on quality flags
contained in the file (``l2_flags``). This option defaults to ``False``, but
when set to ``True`` the "CHLWARN" pixels of the ``l2_flags`` variable
are masked out. These pixels represent data where the chlorophyll algorithm
warned about the quality of the result.

"""

import datetime as dt

from satpy.readers.core.hdf4 import HDF4FileHandler
from satpy.readers.core.netcdf import NetCDF4FileHandler


class _SEADASL2Base:
    """Simple handler of SEADAS L2 files."""

    def __init__(self, filename, filename_info, filetype_info, apply_quality_flags=False):
        """Initialize file handler and determine if data quality flags should be applied."""
        super().__init__(filename, filename_info, filetype_info)
        self.apply_quality_flags = apply_quality_flags and self.l2_flags_var_name in self

    def _add_satpy_metadata(self, data):
        data.attrs["sensor"] = self.sensor_names
        data.attrs["platform_name"] = self._platform_name()
        data.attrs["rows_per_scan"] = self._rows_per_scan()
        return data

    def _rows_per_scan(self):
        if "modis" in self.sensor_names:
            return 10
        if "viirs" in self.sensor_names:
            return 16
        if "oci" in self.sensor_names:
            return 0
        raise ValueError(f"Don't know how to read data for sensors: {self.sensor_names}")

    def _platform_name(self):
        platform = self[self.platform_attr_name]
        platform_dict = {"NPP": "Suomi-NPP",
                         "JPSS-1": "NOAA-20",
                         "JPSS-2": "NOAA-21"}
        return platform_dict.get(platform, platform)

    @property
    def start_time(self):
        """Get the starting observation time of this file's data."""
        start_time = self[self.start_time_attr_name]
        return dt.datetime.strptime(start_time[:-3], self.time_format)

    @property
    def end_time(self):
        """Get the ending observation time of this file's data."""
        end_time = self[self.end_time_attr_name]
        return dt.datetime.strptime(end_time[:-3], self.time_format)

    @property
    def sensor_names(self):
        """Get sensor for the current file's data."""
        # Example: MODISA or VIIRSN or VIIRSJ1
        sensor_name = self[self.sensor_attr_name].lower()
        if sensor_name.startswith("modis"):
            return {"modis"}
        if sensor_name.startswith("viirs"):
            return {"viirs"}
        # Example: OCI
        return {sensor_name}

    def get_dataset(self, data_id, dataset_info):
        """Get DataArray for the specified DataID."""
        file_key, data = self._get_file_key_and_variable(data_id, dataset_info)
        data = self._filter_by_valid_min_max(data)
        data = self._rename_2d_dims_if_necessary(data)
        data = self._mask_based_on_l2_flags(data)
        for attr_name in ("standard_name", "long_name", "units"):
            val = data.attrs[attr_name]
            if val[-1] == "\x00":
                data.attrs[attr_name] = data.attrs[attr_name][:-1]
        data = self._add_satpy_metadata(data)
        return data

    def _get_file_key_and_variable(self, data_id, dataset_info):
        file_keys = dataset_info.get("file_key", data_id["name"])
        if not isinstance(file_keys, list):
            file_keys = [file_keys]
        for file_key in file_keys:
            try:
                data = self[file_key]
                return file_key, data
            except KeyError:
                continue
        raise KeyError(f"Unable to find any of the possible keys for {data_id}: {file_keys}")

    def _rename_2d_dims_if_necessary(self, data_arr):
        if data_arr.ndim != 2 or data_arr.dims == ("y", "x"):
            return data_arr
        return data_arr.rename(dict(zip(data_arr.dims, ("y", "x"))))

    def _filter_by_valid_min_max(self, data_arr):
        valid_range = self._valid_min_max(data_arr)
        data_arr = data_arr.where(valid_range[0] <= data_arr)
        data_arr = data_arr.where(data_arr <= valid_range[1])
        return data_arr

    def _valid_min_max(self, data_arr):
        try:
            return data_arr.attrs["valid_range"]
        except KeyError:
            return data_arr.attrs["valid_min"], data_arr.attrs["valid_max"]

    def _mask_based_on_l2_flags(self, data_arr):
        standard_name = data_arr.attrs.get("standard_name", "")
        if self.apply_quality_flags and not ("lon" in standard_name or "lat" in standard_name):
            l2_flags = self[self.l2_flags_var_name]
            l2_flags = self._rename_2d_dims_if_necessary(l2_flags)
            mask = (l2_flags & 0b00000000010000000000000000000000) != 0
            data_arr = data_arr.where(~mask)
        return data_arr


class SEADASL2NetCDFFileHandler(_SEADASL2Base, NetCDF4FileHandler):
    """Simple handler of SEADAS L2 NetCDF4 files."""

    start_time_attr_name = "/attr/time_coverage_start"
    end_time_attr_name = "/attr/time_coverage_end"
    time_format = "%Y-%m-%dT%H:%M:%S.%f"
    platform_attr_name = "/attr/platform"
    sensor_attr_name = "/attr/instrument"
    l2_flags_var_name = "geophysical_data/l2_flags"


class SEADASL2HDFFileHandler(_SEADASL2Base, HDF4FileHandler):
    """Simple handler of SEADAS L2 HDF4 files."""

    start_time_attr_name = "/attr/Start Time"
    end_time_attr_name = "/attr/End Time"
    time_format = "%Y%j%H%M%S"
    platform_attr_name = "/attr/Mission"
    sensor_attr_name = "/attr/Sensor Name"
    l2_flags_var_name = "l2_flags"



"""Set CF-compliant spatial and temporal coordinates."""
from __future__ import annotations

import logging
import warnings
from collections import defaultdict
from contextlib import suppress

import numpy as np
import xarray as xr
from dask.base import tokenize
from pyproj import CRS
from pyresample.geometry import AreaDefinition, SwathDefinition

logger = logging.getLogger(__name__)


EPOCH = u"seconds since 1970-01-01 00:00:00"


def add_xy_coords_attrs(data_arr: xr.DataArray) -> xr.DataArray:
    """Add relevant attributes to x, y coordinates."""
    # If there are no coords, return dataarray
    if not data_arr.coords.keys() & {"x", "y", "crs"}:
        return data_arr
    # If projected area
    if _is_projected(data_arr):
        data_arr = _add_xy_projected_coords_attrs(data_arr)
    else:
        data_arr = _add_xy_geographic_coords_attrs(data_arr)
    if "crs" in data_arr.coords:
        data_arr = data_arr.drop_vars("crs")
    return data_arr


def _is_projected(data_arr: xr.DataArray) -> bool:
    """Guess whether data are projected or not."""
    crs = _try_to_get_crs(data_arr)
    if crs:
        return crs.is_projected
    units = _try_get_units_from_coords(data_arr)
    if units:
        if units.endswith("m"):
            return True
        if units.startswith("degrees"):
            return False
    logger.warning("Failed to tell if data are projected. Assuming yes.")
    return True


def _is_area(data_arr: xr.DataArray) -> bool:
     return isinstance(data_arr.attrs["area"], AreaDefinition)


def _is_swath(data_arr: xr.DataArray) -> bool:
     return isinstance(data_arr.attrs["area"], SwathDefinition)


def _try_to_get_crs(data_arr: xr.DataArray) -> CRS:
    """Try to get a CRS from attributes."""
    if "area" in data_arr.attrs:
        if _is_area(data_arr):
            return data_arr.attrs["area"].crs
        if not _is_swath(data_arr):
            logger.warning(
                f"Could not tell CRS from area of type {type(data_arr.attrs['area']).__name__:s}. "
                "Assuming projected CRS.")
    if "crs" in data_arr.coords:
        return data_arr.coords["crs"].item()


def _try_get_units_from_coords(data_arr: xr.DataArray) -> str | None:
    """Try to retrieve coordinate x/y units."""
    for c in ["x", "y"]:
        with suppress(KeyError):
            # If the data has only 1 dimension, it has only one of x or y coords
            if "units" in data_arr.coords[c].attrs:
                return data_arr.coords[c].attrs["units"]
    return None


def _add_xy_projected_coords_attrs(data_arr: xr.DataArray, x: str = "x", y: str = "y") -> xr.DataArray:
    """Add relevant attributes to x, y coordinates of a projected CRS."""
    if x in data_arr.coords:
        data_arr[x].attrs["standard_name"] = "projection_x_coordinate"
        data_arr[x].attrs["units"] = "m"
    if y in data_arr.coords:
        data_arr[y].attrs["standard_name"] = "projection_y_coordinate"
        data_arr[y].attrs["units"] = "m"
    return data_arr


def _add_xy_geographic_coords_attrs(data_arr: xr.DataArray, x: str = "x", y: str = "y") -> xr.DataArray:
    """Add relevant attributes to x, y coordinates of a geographic CRS."""
    if x in data_arr.coords:
        data_arr[x].attrs["standard_name"] = "longitude"
        data_arr[x].attrs["units"] = "degrees_east"
    if y in data_arr.coords:
        data_arr[y].attrs["standard_name"] = "latitude"
        data_arr[y].attrs["units"] = "degrees_north"
    return data_arr


def set_cf_time_info(data_arr: xr.DataArray, epoch: str | None) -> xr.DataArray:
    """Set CF time attributes and encoding.

    It expand the DataArray with a time dimension if does not yet exists.

    The function assumes

        - that x and y dimensions have at least shape > 1
        - the time coordinate has size 1

    """
    if epoch is None:
        epoch = EPOCH

    data_arr["time"].encoding["units"] = epoch
    data_arr["time"].attrs["standard_name"] = "time"
    data_arr["time"].attrs.pop("bounds", None)

    if "time" not in data_arr.dims and data_arr["time"].size not in data_arr.shape:
        data_arr = data_arr.expand_dims("time")

    return data_arr


def has_projection_coords(data_arrays: dict[str, xr.DataArray]) -> bool:
    """Check if DataArray collection has a "longitude" or "latitude" DataArray."""
    return any(_is_lon_or_lat_dataarray(data_arr) for data_arr in data_arrays.values())


def _is_lon_or_lat_dataarray(data_arr: xr.DataArray) -> bool:
    """Check if the DataArray represents the latitude or longitude coordinate."""
    return data_arr.attrs.get("standard_name", "") in ("longitude", "latitude")


def _get_is_nondimensional_coords_dict(data_arrays: dict[str, xr.DataArray]) -> dict[str, bool]:
    tokens = defaultdict(set)
    for data_arr in data_arrays.values():
        for coord_name in data_arr.coords:
            if not _is_lon_or_lat_dataarray(data_arr[coord_name]) and coord_name not in data_arr.dims:
                tokens[coord_name].add(tokenize(data_arr[coord_name].data))
    return dict([(coord_name, len(tokens) == 1) for coord_name, tokens in tokens.items()])


def _warn_if_pretty_but_not_unique(pretty, coord_name):
    """Warn if coordinates cannot be pretty-formatted due to non-uniqueness."""
    if pretty:
        warnings.warn(
            f'Cannot pretty-format "{coord_name}" coordinates because they are '
            'not identical among the given datasets',
            stacklevel=2
        )


def _rename_coords(data_arrays: dict[str, xr.DataArray], coord_name: str) -> dict[str, xr.DataArray]:
    """Rename coordinates in the datasets."""
    for name, dataarray in data_arrays.items():
        if coord_name in dataarray.coords:
            rename = {coord_name: f"{name}_{coord_name}"}
            data_arrays[name] = dataarray.rename(rename)
    return data_arrays


def ensure_unique_nondimensional_coords(
        data_arrays: dict[str, xr.DataArray],
        pretty: bool = False
) -> dict[str, xr.DataArray]:
    """Make non-dimensional coordinates unique among all datasets.

    Non-dimensional coordinates, such as scanline timestamps,
    may occur in multiple datasets with the same name and dimension
    but different values.

    In order to avoid conflicts, prepend the dataset name to the coordinate name.
    If a non-dimensional coordinate is unique among all datasets and ``pretty=True``,
    its name will not be modified.

    Since all datasets must have the same projection coordinates,
    this is not applied to latitude and longitude.

    Args:
        data_arrays:
            Dictionary of (dataset name, dataset)
        pretty:
            Don't modify coordinate names, if possible. Makes the file prettier, but possibly less consistent.

    Returns:
        Dictionary holding the updated datasets

    """
    # Determine which non-dimensional coordinates are unique
    # - coords_unique has structure: {coord_name: True/False}
    is_coords_unique_dict = _get_is_nondimensional_coords_dict(data_arrays)

    # Prepend dataset name, if not unique or no pretty-format desired
    new_dict_dataarrays = data_arrays.copy()
    for coord_name, unique in is_coords_unique_dict.items():
        if not pretty or not unique:
            _warn_if_pretty_but_not_unique(pretty, coord_name)
            new_dict_dataarrays = _rename_coords(new_dict_dataarrays, coord_name)
    return new_dict_dataarrays


def check_unique_projection_coords(data_arrays: dict[str, xr.DataArray]) -> None:
    """Check that all datasets share the same projection coordinates x/y."""
    unique_x = set()
    unique_y = set()
    for dataarray in data_arrays.values():
        if "y" in dataarray.dims:
            token_y = tokenize(dataarray["y"].data)
            unique_y.add(token_y)
        if "x" in dataarray.dims:
            token_x = tokenize(dataarray["x"].data)
            unique_x.add(token_x)
    if len(unique_x) > 1 or len(unique_y) > 1:
        raise ValueError("Datasets to be saved in one file (or one group) must have identical projection coordinates."
                         "Please group them by area or save them in separate files.")


def add_coordinates_attrs_coords(data_arrays: dict[str, xr.DataArray]) -> dict[str, xr.DataArray]:
    """Add to DataArrays the coordinates specified in the 'coordinates' attribute.

    It deal with the 'coordinates' attributes indicating lat/lon coords
    The 'coordinates' attribute is dropped from each DataArray

    If the `coordinates` attribute of a data array links to other dataarrays in the scene, for example
    `coordinates='lon lat'`, add them as coordinates to the data array and drop that attribute.

    In the final call to `xr.Dataset.to_netcdf()` all coordinate relations will be resolved
    and the `coordinates` attributes be set automatically.
    """
    for dataarray_name in data_arrays.keys():
        data_arrays = _add_declared_coordinates(data_arrays,
                                                dataarray_name=dataarray_name)
        # Drop 'coordinates' attribute in any case to avoid conflicts in xr.Dataset.to_netcdf()
        data_arrays[dataarray_name].attrs.pop("coordinates", None)
    return data_arrays


def _add_declared_coordinates(data_arrays: dict[str, xr.DataArray], dataarray_name: str) -> dict[str, xr.DataArray]:
    """Add declared coordinates to the dataarray if they exist."""
    dataarray = data_arrays[dataarray_name]
    declared_coordinates = _get_coordinates_list(dataarray)
    for coord in declared_coordinates:
        if coord not in dataarray.coords:
            data_arrays = _try_add_coordinate(data_arrays,
                                              dataarray_name=dataarray_name,
                                              coord=coord)
    return data_arrays


def _try_add_coordinate(
        data_arrays: dict[str, xr.DataArray],
        dataarray_name: str,
        coord: str
) -> dict[str, xr.DataArray]:
    """Try to add a coordinate to the dataarray, warn if not possible."""
    try:
        dataarray_dims = set(data_arrays[dataarray_name].dims)
        coordinate_dims = set(data_arrays[coord].dims)
        dimensions_to_squeeze = list(coordinate_dims - dataarray_dims)
        data_arrays[dataarray_name][coord] = data_arrays[coord].squeeze(dimensions_to_squeeze, drop=True)
    except KeyError:
        warnings.warn(
            f'Coordinate "{coord}" referenced by dataarray {dataarray_name} does not '
            'exist, dropping reference.',
            stacklevel=2
        )
    return data_arrays


def _get_coordinates_list(data_arr: xr.DataArray) -> list[str]:
    """Return a list with the coordinates names specified in the 'coordinates' attribute."""
    declared_coordinates = data_arr.attrs.get("coordinates", [])
    if isinstance(declared_coordinates, str):
        declared_coordinates = declared_coordinates.split(" ")
    return declared_coordinates


def add_time_bounds_dimension(ds: xr.Dataset, time: str = "time") -> xr.Dataset:
    """Add time bound dimension to xr.Dataset."""
    tbcalc = _TimeBoundsCalculator(ds, time)
    time_bounds = tbcalc.get_time_bounds()
    ds["time_bnds"] = time_bounds
    ds[time].attrs["bounds"] = "time_bnds"
    ds[time].attrs["standard_name"] = "time"
    return ds


class _TimeBoundsCalculator:
    def __init__(self, ds: xr.Dataset, time_dim: str) -> None:
        """Initialize time bounds calculator.

        Args:
            ds: Dataset to be analyzed
            time_dim: Name of time dimension
        """
        self.ds = ds
        self.time_dim = time_dim
        self.times = self.ds[self.time_dim].values

    def get_time_bounds(self) -> xr.DataArray:
        """Get time bounds for the given dataset."""
        if self.times.size > 1:
            bounds = self._get_bounds_multiple_timesteps()
        else:
            bounds = self._get_bounds_single_timestep()
        return xr.DataArray(bounds, dims=[self.time_dim, "bnds_1d"])

    def _get_bounds_multiple_timesteps(self):
        """Get time bounds for a dataset with multiple timesteps.

        Computes bounds based on time coordinates and for the last
        timestep uses the "end_time" attribute. Example for 3
        timesteps::

            bounds = [
                (t[0], t[1]),
                (t[1], t[2]),
                (t[2], max_end_time_from_attrs)
            ]

        """
        bounds = self._get_bounds_for_all_but_last_timestep()
        bounds.append(self._get_bounds_for_last_timestep())
        return bounds

    def _get_bounds_for_all_but_last_timestep(self):
        return [
            [self.times[i], self.times[i + 1]]
            for i in range(len(self.times) - 1)
        ]

    def _get_bounds_for_last_timestep(self):
        return [
            self.times[-1],
            self._get_time_from_attrs("end_time", max)
        ]

    def _get_time_from_attrs(self, time_attr, min_or_max):
        times = [
            data_array.attrs.get(time_attr, None)
            for data_array in self.ds.data_vars.values()
        ]
        time = min_or_max(
            t for t in times if t is not None
        )
        return np.datetime64(time, "ns")

    def _get_bounds_single_timestep(self):
        """Get time bounds for a dataset with a single timestep.

        Computes time bounds entirely from dataset attributes
        start_time and end_time.
        """
        return [
            [self._get_time_from_attrs("start_time", min),
             self._get_time_from_attrs("end_time", max)]
        ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""SAFE SAR L2 OCN format reader.

The OCN data contains various parameters, but mainly the wind speed and direction
calculated from SAR data and input model data from ECMWF

Implemented in this reader is the OWI, Ocean Wind field.

See more at ESA webpage https://sentinel.esa.int/web/sentinel/ocean-wind-field-component
"""

import logging

import numpy as np
import xarray as xr

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import get_legacy_chunk_size

logger = logging.getLogger(__name__)
CHUNK_SIZE = get_legacy_chunk_size()


class SAFENC(BaseFileHandler):
    """Measurement file reader."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init the file reader."""
        super(SAFENC, self).__init__(filename, filename_info,
                                     filetype_info)

        self._start_time = filename_info["start_time"]
        self._end_time = filename_info["end_time"]
        # For some SAFE packages, fstart_time differs, but start_time is the same
        # To avoid over writing exiting file with same start_time, a solution is to
        # use fstart_time
        self._fstart_time = filename_info["fstart_time"]
        self._fend_time = filename_info["fend_time"]

        self._polarization = filename_info["polarization"]

        self.lats = None
        self.lons = None
        self._shape = None
        self.area = None

        self.nc = xr.open_dataset(filename,
                                  decode_cf=True,
                                  mask_and_scale=False,
                                  chunks={"owiAzSize": CHUNK_SIZE,
                                          "owiRaSize": CHUNK_SIZE})
        self.nc = self.nc.rename({"owiAzSize": "y"})
        self.nc = self.nc.rename({"owiRaSize": "x"})
        self.filename = filename

    def get_dataset(self, key, info):
        """Load a dataset."""
        if key["name"] in ["owiLat", "owiLon"]:
            if self.lons is None or self.lats is None:
                self.lons = self.nc["owiLon"]
                self.lats = self.nc["owiLat"]
            if key["name"] == "owiLat":
                res = self.lats
            else:
                res = self.lons
            res.attrs = info
        else:
            res = self._get_data_channels(key, info)

        if "missionName" in self.nc.attrs:
            res.attrs.update({"platform_name": self.nc.attrs["missionName"]})

        res.attrs.update({"fstart_time": self._fstart_time})
        res.attrs.update({"fend_time": self._fend_time})

        if not self._shape:
            self._shape = res.shape

        return res

    def _get_data_channels(self, key, info):
        res = self.nc[key["name"]]
        if key["name"] in ["owiHs", "owiWl", "owiDirmet"]:
            res = xr.DataArray(res, dims=["y", "x", "oswPartitions"])
        elif key["name"] in ["owiNrcs", "owiNesz", "owiNrcsNeszCorr"]:
            res = xr.DataArray(res, dims=["y", "x", "oswPolarisation"])
        elif key["name"] in ["owiPolarisationName"]:
            res = xr.DataArray(res, dims=["owiPolarisation"])
        elif key["name"] in ["owiCalConstObsi", "owiCalConstInci"]:
            res = xr.DataArray(res, dims=["owiIncSize"])
        elif key["name"].startswith("owi"):
            res = xr.DataArray(res, dims=["y", "x"])
        else:
            res = xr.DataArray(res, dims=["y", "x"])
        res.attrs.update(info)
        if "_FillValue" in res.attrs:
            res = res.where(res != res.attrs["_FillValue"])
            res.attrs["_FillValue"] = np.nan
        return res

    @property
    def start_time(self):
        """Product start_time, parsed from the measurement file name."""
        return self._start_time

    @property
    def end_time(self):
        """Product end_time, parsed from the measurement file name."""
        return self._end_time

    @property
    def fstart_time(self):
        """Product fstart_time meaning the start time parsed from the SAFE directory."""
        return self._fstart_time

    @property
    def fend_time(self):
        """Product fend_time meaning the end time parsed from the SAFE directory."""
        return self._fend_time



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Interface to GEOCAT HDF4 or NetCDF4 products.

Note: GEOCAT files do not currently have projection information or precise
pixel resolution information. Additionally the longitude and latitude arrays
are stored as 16-bit integers which causes loss of precision. For this reason
the lon/lats can't be used as a reliable coordinate system to calculate the
projection X/Y coordinates.

Until GEOCAT adds projection information and X/Y coordinate arrays, this
reader will estimate the geostationary area the best it can. It currently
takes a single lon/lat point as reference and uses hardcoded resolution
and projection information to calculate the area extents.

"""
from __future__ import annotations

import logging

import numpy as np
from pyproj import Proj
from pyresample import geometry

from satpy.readers.core.netcdf import NetCDF4FileHandler, netCDF4

LOG = logging.getLogger(__name__)


CF_UNITS = {
    "none": "1",
}

# GEOCAT currently doesn't include projection information in it's files
GEO_PROJS = {
    "GOES-16": "+proj=geos +lon_0={lon_0:0.02f} +h=35786023.0 +a=6378137.0 +b=6356752.31414 +sweep=x +units=m +no_defs",
    "GOES-17": "+proj=geos +lon_0={lon_0:0.02f} +h=35786023.0 +a=6378137.0 +b=6356752.31414 +sweep=x +units=m +no_defs",
    "HIMAWARI-8": "+proj=geos +over +lon_0=140.7 +h=35785863 +a=6378137 +b=6356752.299581327 +units=m +no_defs",
}


class GEOCATFileHandler(NetCDF4FileHandler):
    """GEOCAT netCDF4 file handler.

    **Loading data with decode_times=True**

    By default, this reader will use ``xarray_kwargs={"engine": "netcdf4", "decode_times": False}``.
    to match behavior of xarray when the geocat reader was first written.  To use different options
    use reader_kwargs when loading the Scene::

        scene = satpy.Scene(filenames,
                            reader='geocat',
                            reader_kwargs={'xarray_kwargs': {'engine': 'netcdf4', 'decode_times': True}})
    """

    def __init__(self, filename, filename_info, filetype_info,
                 **kwargs):
        """Open and perform initial investigation of NetCDF file."""
        kwargs.setdefault("xarray_kwargs", {}).setdefault(
            "engine", "netcdf4")
        kwargs.setdefault("xarray_kwargs", {}).setdefault(
            "decode_times", False)

        super(GEOCATFileHandler, self).__init__(
            filename, filename_info, filetype_info,
            xarray_kwargs=kwargs["xarray_kwargs"])

    sensors = {
        "goes": "goes_imager",
        "himawari8": "ahi",
        "goes16": "abi",  # untested
        "goesr": "abi",  # untested
    }
    platforms: dict[str, str] = {
    }
    resolutions = {
        "abi": {
            1: 1002.0086577437705,
            2: 2004.0173154875411,
        },
        "ahi": {
            1: 999.9999820317674,  # assumption
            2: 1999.999964063535,
            4: 3999.99992812707,
        }
    }

    def get_sensor(self, sensor):
        """Get sensor."""
        last_resort = None
        for k, v in self.sensors.items():
            if k == sensor:
                return v
            if k in sensor:
                last_resort = v
        if last_resort:
            return last_resort
        raise ValueError("Unknown sensor '{}'".format(sensor))

    def get_platform(self, platform):
        """Get platform."""
        for k, v in self.platforms.items():
            if k in platform:
                return v
        return platform

    def _get_proj(self, platform, ref_lon):
        if platform == "GOES-16" and -76. < ref_lon < -74.:
            # geocat file holds the *actual* subsatellite point, not the
            # projection (-75.2 actual versus -75 projection)
            ref_lon = -75.
        return GEO_PROJS[platform].format(lon_0=ref_lon)

    @property
    def sensor_names(self):
        """Get sensor names."""
        return [self.get_sensor(self["/attr/Sensor_Name"])]

    @property
    def start_time(self):
        """Get start time."""
        return self.filename_info["start_time"]

    @property
    def end_time(self):
        """Get end time."""
        return self.filename_info.get("end_time", self.start_time)

    @property
    def is_geo(self):
        """Check platform."""
        platform = self.get_platform(self["/attr/Platform_Name"])
        return platform in GEO_PROJS

    @property
    def resolution(self):
        """Get resolution."""
        elem_res = self["/attr/Element_Resolution"]
        return int(elem_res * 1000)

    def _calc_area_resolution(self, ds_res):
        elem_res = round(ds_res / 1000.)  # mimic 'Element_Resolution' attribute from above
        sensor = self.get_sensor(self["/attr/Sensor_Name"])
        return self.resolutions.get(sensor, {}).get(int(elem_res),
                                                    elem_res * 1000.)

    def available_datasets(self, configured_datasets=None):
        """Update information for or add datasets provided by this file.

        If this file handler can load a dataset then it will supplement the
        dataset info with the resolution and possibly coordinate datasets
        needed to load it. Otherwise it will continue passing the dataset
        information down the chain.

        See
        :meth:`satpy.readers.core.file_handlers.BaseFileHandler.available_datasets`
        for details.

        """
        res = self.resolution
        coordinates = ("pixel_longitude", "pixel_latitude")
        handled_variables = set()

        # update previously configured datasets
        for is_avail, ds_info in (configured_datasets or []):
            this_res = ds_info.get("resolution")
            this_coords = ds_info.get("coordinates")
            # some other file handler knows how to load this
            if is_avail is not None:
                yield is_avail, ds_info

            var_name = ds_info.get("file_key", ds_info["name"])
            matches = self.file_type_matches(ds_info["file_type"])
            # we can confidently say that we can provide this dataset and can
            # provide more info
            if matches and var_name in self and this_res != res:
                handled_variables.add(var_name)
                new_info = ds_info.copy()  # don't mess up the above yielded
                new_info["resolution"] = res
                if not self.is_geo and this_coords is None:
                    new_info["coordinates"] = coordinates
                yield True, new_info
            elif is_avail is None:
                # if we didn't know how to handle this dataset and no one else did
                # then we should keep it going down the chain
                yield is_avail, ds_info

        # Provide new datasets
        for var_name, val in self.file_content.items():
            if var_name in handled_variables:
                continue
            if isinstance(val, netCDF4.Variable):
                ds_info = {
                    "file_type": self.filetype_info["file_type"],
                    "resolution": res,
                    "name": var_name,
                }
                if not self.is_geo:
                    ds_info["coordinates"] = coordinates
                yield True, ds_info

    def get_shape(self, dataset_id, ds_info):
        """Get shape."""
        var_name = ds_info.get("file_key", dataset_id["name"])
        return self[var_name + "/shape"]

    def _first_good_nav(self, lon_arr, lat_arr):
        if hasattr(lon_arr, "mask"):
            good_indexes = np.nonzero(~lon_arr.mask)
        else:
            # no masked values found in auto maskandscale
            good_indexes = ([0], [0])
        # nonzero returns (<ndarray of row indexes>, <ndarray of col indexes>)
        return tuple(x[0] for x in good_indexes)

    def _get_extents(self, proj, res, lon_arr, lat_arr):
        p = Proj(proj)
        res = float(res)
        first_good = self._first_good_nav(lon_arr, lat_arr)
        one_x, one_y = p(lon_arr[first_good], lat_arr[first_good])
        left_x = one_x - res * first_good[1]
        right_x = left_x + res * lon_arr.shape[1]
        top_y = one_y + res * first_good[0]
        bot_y = top_y - res * lon_arr.shape[0]
        half_x = res / 2.
        half_y = res / 2.
        return (left_x - half_x,
                bot_y - half_y,
                right_x + half_x,
                top_y + half_y)

    def _load_nav(self, name):
        nav = self[name]
        factor = self[name + "/attr/scale_factor"]
        offset = self[name + "/attr/add_offset"]
        fill = self[name + "/attr/_FillValue"]
        nav = nav[:]
        mask = nav == fill
        nav = np.ma.masked_array(nav * factor + offset, mask=mask)
        return nav[:]

    def get_area_def(self, dsid):
        """Get area definition."""
        if not self.is_geo:
            raise NotImplementedError("Don't know how to get the Area Definition for this file")

        platform = self.get_platform(self["/attr/Platform_Name"])
        res = self._calc_area_resolution(dsid["resolution"])
        proj = self._get_proj(platform, float(self["/attr/Subsatellite_Longitude"]))
        area_name = "{} {} Area at {}m".format(
            platform,
            self.metadata.get("sector_id", ""),
            int(res))
        lon = self._load_nav("pixel_longitude")
        lat = self._load_nav("pixel_latitude")
        extents = self._get_extents(proj, res, lon, lat)
        area_def = geometry.AreaDefinition(
            area_name,
            area_name,
            area_name,
            proj,
            lon.shape[1],
            lon.shape[0],
            area_extent=extents,
        )
        return area_def

    def get_metadata(self, dataset_id, ds_info):
        """Get metadata."""
        var_name = ds_info.get("file_key", dataset_id["name"])
        shape = self.get_shape(dataset_id, ds_info)
        info = getattr(self[var_name], "attrs", {})
        info["shape"] = shape
        info.update(ds_info)
        u = info.get("units")
        if u in CF_UNITS:
            # CF compliance
            info["units"] = CF_UNITS[u]

        info["sensor"] = self.get_sensor(self["/attr/Sensor_Name"])
        info["platform_name"] = self.get_platform(self["/attr/Platform_Name"])
        info["resolution"] = dataset_id["resolution"]
        if var_name == "pixel_longitude":
            info["standard_name"] = "longitude"
        elif var_name == "pixel_latitude":
            info["standard_name"] = "latitude"

        return info

    def get_dataset(self, dataset_id, ds_info):
        """Get dataset."""
        var_name = ds_info.get("file_key", dataset_id["name"])
        # FUTURE: Metadata retrieval may be separate
        info = self.get_metadata(dataset_id, ds_info)
        data = self[var_name]
        fill = self[var_name + "/attr/_FillValue"]
        factor = self.get(var_name + "/attr/scale_factor")
        offset = self.get(var_name + "/attr/add_offset")
        valid_range = self.get(var_name + "/attr/valid_range")

        data = data.where(data != fill)
        if valid_range is not None:
            data = data.where((data >= valid_range[0]) & (data <= valid_range[1]))
        if factor is not None and offset is not None:
            data = data * factor + offset

        data.attrs.update(info)
        data = data.rename({"lines": "y", "elements": "x"})
        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2009-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Satpy Package initializer."""

try:
    from satpy.version import version as __version__  # noqa
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "No module named satpy.version. This could mean "
        "you didn't install 'satpy' properly. Try reinstalling ('pip "
        "install').")

from satpy._config import config  # noqa
from satpy.dataset import DataID, DataQuery  # noqa
from satpy.dataset.data_dict import DatasetDict  # noqa
from satpy.multiscene import MultiScene  # noqa
from satpy.readers.core.config import available_readers  # noqa
from satpy.readers.core.grouping import find_files_and_readers  # noqa
from satpy.scene import Scene  # noqa
from satpy.utils import get_logger  # noqa
from satpy.writers.core.config import available_writers  # noqa

log = get_logger("satpy")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The HRIT msg reader tests package."""

import datetime as dt
from unittest import mock

import numpy as np
import pytest
import xarray as xr

from satpy.readers.seviri_l1b_nc import NCSEVIRIFileHandler
from satpy.tests.reader_tests.test_seviri_base import ORBIT_POLYNOMIALS
from satpy.tests.reader_tests.test_seviri_l1b_calibration import TestFileHandlerCalibrationBase
from satpy.tests.utils import assert_attrs_equal, make_dataid

channel_keys_dict = {"VIS006": "ch1", "IR_108": "ch9"}


def to_cds_time(time):
    """Convert datetime to (days, msecs) since 1958-01-01."""
    if isinstance(time, dt.datetime):
        time = np.datetime64(time)
    t0 = np.datetime64("1958-01-01 00:00")
    delta = time - t0
    days = (delta / np.timedelta64(1, "D")).astype(int)
    msecs = delta / np.timedelta64(1, "ms") - days * 24 * 3600 * 1E3
    return days, msecs


class TestNCSEVIRIFileHandler(TestFileHandlerCalibrationBase):
    """Unit tests for SEVIRI netCDF reader."""

    def _get_fake_dataset(self, counts, h5netcdf):
        """Create a fake dataset.

        Args:
            counts (xr.DataArray):
                Array with data.
            h5netcdf (boolean):
                If True an array attribute will be created which is common
                for the h5netcdf backend in xarray for scalar values.

        """
        acq_time_day = np.repeat([1, 1], 11).reshape(2, 11)
        acq_time_msec = np.repeat([1000, 2000], 11).reshape(2, 11)
        line_validity = np.repeat([3, 3], 11).reshape(2, 11)
        line_geom_radio_quality = np.repeat([4, 4], 11).reshape(2, 11)
        orbit_poly_start_day, orbit_poly_start_msec = to_cds_time(
            np.array([dt.datetime(2019, 12, 31, 18),
                      dt.datetime(2019, 12, 31, 22)],
                     dtype="datetime64")
        )
        orbit_poly_end_day, orbit_poly_end_msec = to_cds_time(
            np.array([dt.datetime(2019, 12, 31, 22),
                      dt.datetime(2020, 1, 1, 2)],
                     dtype="datetime64")
        )
        counts = counts.rename({
            "y": "num_rows_vis_ir",
            "x": "num_columns_vis_ir"
        })
        scan_time_days, scan_time_msecs = to_cds_time(self.scan_time)
        ds = xr.Dataset(
            {
                "ch1": counts.copy(),
                "ch9": counts.copy(),
                "HRV": (("num_rows_hrv", "num_columns_hrv"), [[1, 2, 3],
                                                              [4, 5, 6],
                                                              [7, 8, 9]]),
                "planned_chan_processing": self.radiance_types,
                "channel_data_visir_data_l10_line_mean_acquisition_time_day": (
                    ("num_rows_vis_ir", "channels_vis_ir_dim"),
                    acq_time_day
                ),
                "channel_data_visir_data_l10_line_mean_acquisition_msec": (
                    ("num_rows_vis_ir", "channels_vis_ir_dim"),
                    acq_time_msec
                ),
                "channel_data_visir_data_line_validity": (
                    ("num_rows_vis_ir", "channels_vis_ir_dim"),
                    line_validity
                ),
                "channel_data_visir_data_line_geometric_quality": (
                    ("num_rows_vis_ir", "channels_vis_ir_dim"),
                    line_geom_radio_quality
                ),
                "channel_data_visir_data_line_radiometric_quality": (
                    ("num_rows_vis_ir", "channels_vis_ir_dim"),
                    line_geom_radio_quality
                ),
                "orbit_polynomial_x": (
                    ("orbit_polynomial_dim_row",
                     "orbit_polynomial_dim_col"),
                    ORBIT_POLYNOMIALS["X"][0:2]
                ),
                "orbit_polynomial_y": (
                    ("orbit_polynomial_dim_row",
                     "orbit_polynomial_dim_col"),
                    ORBIT_POLYNOMIALS["Y"][0:2]
                ),
                "orbit_polynomial_z": (
                    ("orbit_polynomial_dim_row",
                     "orbit_polynomial_dim_col"),
                    ORBIT_POLYNOMIALS["Z"][0:2]
                ),
                "orbit_polynomial_start_time_day": (
                    "orbit_polynomial_dim_row",
                    orbit_poly_start_day
                ),
                "orbit_polynomial_start_time_msec": (
                    "orbit_polynomial_dim_row",
                    orbit_poly_start_msec
                ),
                "orbit_polynomial_end_time_day": (
                    "orbit_polynomial_dim_row",
                    orbit_poly_end_day
                ),
                "orbit_polynomial_end_time_msec": (
                    "orbit_polynomial_dim_row",
                    orbit_poly_end_msec
                ),
            },
            attrs={
                "equatorial_radius": 6378.169,
                "north_polar_radius": 6356.5838,
                "south_polar_radius": 6356.5838,
                "longitude_of_SSP": 0.0,
                "nominal_longitude": -3.5,
                "satellite_id": self.platform_id,
                "true_repeat_cycle_start_day": scan_time_days,
                "true_repeat_cycle_start_mi_sec": scan_time_msecs,
                "planned_repeat_cycle_end_day": scan_time_days,
                "planned_repeat_cycle_end_mi_sec": scan_time_msecs,
                "north_most_line": 3712,
                "east_most_pixel": 1,
                "west_most_pixel": 3712,
                "south_most_line": 1,
                "vis_ir_grid_origin": 0,
                "vis_ir_column_dir_grid_step": 3.0004032,
                "vis_ir_line_dir_grid_step": 3.0004032,
                "type_of_earth_model": "0x02",
                "nominal_image_scanning": "T",
            }
        )

        if h5netcdf:
            nattrs = {"equatorial_radius": np.array([6378.169]),
                      "north_polar_radius": np.array([6356.5838]),
                      "south_polar_radius": np.array([6356.5838]),
                      "longitude_of_SSP": np.array([0.0]),
                      "vis_ir_column_dir_grid_step": np.array([3.0004032]),
                      "vis_ir_line_dir_grid_step": np.array([3.0004032])
                      }

            ds.attrs.update(nattrs)

        ds["ch1"].attrs.update({
            "scale_factor": self.gains_nominal[0],
            "add_offset": self.offsets_nominal[0]
        })
        # IR_108 is dataset with key ch9
        ds["ch9"].attrs.update({
            "scale_factor": self.gains_nominal[8],
            "add_offset": self.offsets_nominal[8],
        })

        # Add some attributes so that the reader can strip them
        strip_attrs = {
            "comment": None,
            "long_name": None,
            "valid_min": None,
            "valid_max": None
        }
        for name in ["ch1", "ch9"]:
            ds[name].attrs.update(strip_attrs)

        return ds

    @pytest.fixture
    def h5netcdf(self):
        """Fixture for xr backend choice."""
        return False

    @pytest.fixture(name="file_handler")
    def file_handler(self, counts, h5netcdf):
        """Create a mocked file handler."""
        with mock.patch(
            "satpy.readers.seviri_l1b_nc.open_dataset",
            return_value=self._get_fake_dataset(counts=counts, h5netcdf=h5netcdf)
        ):
            return NCSEVIRIFileHandler(
                "filename",
                {"platform_shortname": "MSG3",
                 "start_time": self.scan_time,
                 "service": "MSG"},
                {"filetype": "info"}
            )

    @pytest.mark.parametrize(
        ("channel", "calibration", "use_ext_coefs"),
        [
            # VIS channel, internal coefficients
            ("VIS006", "counts", False),
            ("VIS006", "radiance", False),
            ("VIS006", "reflectance", False),
            # VIS channel, external coefficients
            ("VIS006", "radiance", True),
            ("VIS006", "reflectance", True),
            # IR channel, internal coefficients
            ("IR_108", "counts", False),
            ("IR_108", "radiance", False),
            ("IR_108", "brightness_temperature", False),
            # IR channel, external coefficients
            ("IR_108", "radiance", True),
            ("IR_108", "brightness_temperature", True),
            # FUTURE: Enable once HRV reading has been fixed.
            # # HRV channel, internal coefficiens
            # ('HRV', 'counts', False),
            # ('HRV', 'radiance', False),
            # ('HRV', 'reflectance', False),
            # # HRV channel, external coefficients (mode should have no effect)
            # ('HRV', 'radiance', True),
            # ('HRV', 'reflectance', True),
        ]
    )
    def test_calibrate(
            self, file_handler, channel, calibration, use_ext_coefs
    ):
        """Test the calibration."""
        external_coefs = self.external_coefs if use_ext_coefs else {}
        expected = self._get_expected(
            channel=channel,
            calibration=calibration,
            calib_mode="NOMINAL",
            use_ext_coefs=use_ext_coefs
        )
        fh = file_handler
        fh.ext_calib_coefs = external_coefs
        dataset_id = make_dataid(name=channel, calibration=calibration)

        key = channel_keys_dict[channel]

        res = fh.calibrate(fh.nc[key], dataset_id)
        xr.testing.assert_allclose(res, expected)

    def test_mask_bad_quality(self, file_handler):
        """Test masking of bad quality scan lines."""
        channel = "VIS006"
        key = channel_keys_dict[channel]
        dataset_info = {
            "nc_key": key,
            "units": "units",
            "wavelength": "wavelength",
            "standard_name": "standard_name"
        }
        expected = self._get_expected(
            channel=channel,
            calibration="radiance",
            calib_mode="NOMINAL",
            use_ext_coefs=False
        )

        fh = file_handler

        res = fh._mask_bad_quality(fh.nc[key], dataset_info)
        new_data = np.zeros_like(expected.data).astype("float32")
        new_data[:, :] = np.nan
        expected = expected.copy(data=new_data)
        xr.testing.assert_allclose(res, expected)

    @pytest.mark.parametrize(
        ("channel", "calibration", "mask_bad_quality_scan_lines"),
        [
            ("VIS006", "reflectance", True),
            ("VIS006", "reflectance", False),
            ("IR_108", "brightness_temperature", True)
         ]
    )
    def test_get_dataset(self, file_handler, channel, calibration, mask_bad_quality_scan_lines):
        """Test getting the dataset."""
        dataset_id = make_dataid(name=channel, calibration=calibration)
        key = channel_keys_dict[channel]
        dataset_info = {
            "nc_key": key,
            "units": "units",
            "wavelength": "wavelength",
            "standard_name": "standard_name"
        }

        file_handler.mask_bad_quality_scan_lines = mask_bad_quality_scan_lines
        res = file_handler.get_dataset(dataset_id, dataset_info)

        # Test scanline acquisition times
        expected = self._get_expected(
            channel=channel,
            calibration=calibration,
            calib_mode="NOMINAL",
            use_ext_coefs=False
        )
        expected.attrs = {
            "orbital_parameters": {
                "satellite_actual_longitude": -3.541742131915741,
                "satellite_actual_latitude": -0.5203765167594427,
                "satellite_actual_altitude": 35783419.16135868,
                "satellite_nominal_longitude": -3.5,
                "satellite_nominal_latitude": 0.0,
                "projection_longitude": 0.0,
                "projection_latitude": 0.0,
                "projection_altitude": 35785831.0
            },
            "time_parameters": {
                "nominal_start_time": dt.datetime(2020, 1, 1, 0, 0),
                "nominal_end_time": dt.datetime(2020, 1, 1, 0, 0),
                "observation_start_time": dt.datetime(2020, 1, 1, 0, 0),
                "observation_end_time": dt.datetime(2020, 1, 1, 0, 0),
            },
            "georef_offset_corrected": True,
            "platform_name": "Meteosat-11",
            "sensor": "seviri",
            "units": "units",
            "wavelength": "wavelength",
            "standard_name": "standard_name"
        }
        expected["acq_time"] = ("y", [np.datetime64("1958-01-02 00:00:01").astype("datetime64[ns]"),
                                      np.datetime64("1958-01-02 00:00:02").astype("datetime64[ns]")])
        expected = expected[::-1]  # reader flips data upside down
        if mask_bad_quality_scan_lines:
            expected = file_handler._mask_bad_quality(expected, dataset_info)

        xr.testing.assert_allclose(res, expected)

        for key in ["sun_earth_distance_correction_applied",
                    "sun_earth_distance_correction_factor"]:
            res.attrs.pop(key, None)
        assert_attrs_equal(res.attrs, expected.attrs, tolerance=1e-4)

    def test_time(self, file_handler):
        """Test start/end nominal/observation time handling."""
        assert dt.datetime(2020, 1, 1, 0, 0) == file_handler.observation_start_time
        assert dt.datetime(2020, 1, 1, 0, 0) == file_handler.start_time
        assert file_handler.start_time == file_handler.nominal_start_time

        assert dt.datetime(2020, 1, 1, 0, 0) == file_handler.observation_end_time
        assert file_handler.end_time == file_handler.nominal_end_time
        assert dt.datetime(2020, 1, 1, 0, 0) == file_handler.end_time

    def test_repeat_cycle_duration(self, file_handler):
        """Test repeat cycle handling for FD or ReduscedScan."""
        assert 15 == file_handler._repeat_cycle_duration
        # Change the reducescan scenario to test the repeat cycle duration handling
        file_handler.nc.attrs["nominal_image_scanning"] = ""
        file_handler.nc.attrs["reduced_scanning"] = "T"
        # file_handler.trailer['15TRAILER']['ImageProductionStats']['ActualScanningSummary']['ReducedScan'] = 1
        assert 5 == file_handler._repeat_cycle_duration

    def test_satpos_no_valid_orbit_polynomial(self, file_handler):
        """Test satellite position if there is no valid orbit polynomial."""
        dataset_id = make_dataid(name="VIS006", calibration="counts")
        dataset_info = {
            "name": "VIS006",
            "nc_key": "ch1",
            "units": "units",
            "wavelength": "wavelength",
            "standard_name": "standard_name"
        }
        file_handler.nc["orbit_polynomial_start_time_day"] = 0
        file_handler.nc["orbit_polynomial_end_time_day"] = 0
        with pytest.warns(UserWarning, match=r"No orbit polynomial valid for"):
            res = file_handler.get_dataset(dataset_id, dataset_info)
        assert "satellite_actual_longitude" not in res.attrs[
            "orbital_parameters"]

    @pytest.mark.parametrize("h5netcdf", [True])
    def test_h5netcdf_pecularity(self, file_handler, h5netcdf):
        """Test conversion of attributes when xarray is used with h5netcdf backend."""
        fh = file_handler
        assert isinstance(fh.mda["projection_parameters"]["a"], float)



#!/usr/bin/env python
# Copyright (c) 2018-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for cloud product compositors."""

import unittest

import dask.array as da
import numpy as np
import xarray as xr


class TestCloudCompositorWithoutCloudfree:
    """Test the CloudCompositorWithoutCloudfree."""

    def setup_method(self):
        """Set up the test case."""
        from satpy.composites.cloud_products import CloudCompositorWithoutCloudfree
        self.colormap_composite = CloudCompositorWithoutCloudfree("test_cmap_compositor")

        self.exp = np.array([[4, 3, 2], [2, 3, np.nan], [8, 7, 655350]])
        self.exp_bad_oc = np.array([[4, 3, 2],
                                    [2, np.nan, 4],
                                    [np.nan, 7, 255]])

    def test_call_numpy_with_invalid_value_in_status(self):
        """Test the CloudCompositorWithoutCloudfree composite generation."""
        status = xr.DataArray(np.array([[0, 0, 0], [0, 0, 65535], [0, 0, 1]]), dims=["y", "x"],
                              attrs={"_FillValue": 65535})
        data = xr.DataArray(np.array([[4, 3, 2], [2, 3, np.nan], [8, 7, np.nan]], dtype=np.float32),
                            dims=["y", "x"],
                            attrs={"_FillValue": 65535,
                                   "scaled_FillValue": 655350})
        res = self.colormap_composite([data, status])
        np.testing.assert_allclose(res, self.exp, atol=1e-4)

    def test_call_dask_with_invalid_value_in_status(self):
        """Test the CloudCompositorWithoutCloudfree composite generation."""
        status = xr.DataArray(da.from_array(np.array([[0, 0, 0], [0, 0, 65535], [0, 0, 1]])), dims=["y", "x"],
                              attrs={"_FillValue": 65535})
        data = xr.DataArray(da.from_array(np.array([[4, 3, 2], [2, 3, np.nan], [8, 7, np.nan]], dtype=np.float32)),
                            dims=["y", "x"],
                            attrs={"_FillValue": 99,
                                   "scaled_FillValue": 655350})
        res = self.colormap_composite([data, status])
        np.testing.assert_allclose(res, self.exp, atol=1e-4)

    def test_call_bad_optical_conditions(self):
        """Test the CloudCompositorWithoutCloudfree composite generation."""
        status = xr.DataArray(da.from_array(np.array([[0, 0, 0], [3, 3, 3], [0, 0, 1]])), dims=["y", "x"],
                              attrs={"_FillValue": 65535,
                                     "flag_meanings": "bad_optical_conditions"})
        data = xr.DataArray(np.array([[4, 3, 2], [2, 255, 4], [255, 7, 255]], dtype=np.uint8),
                            dims=["y", "x"],
                            name="cmic_cre",
                            attrs={"_FillValue": 255,
                                   "scaled_FillValue": 255})
        res = self.colormap_composite([data, status])
        np.testing.assert_allclose(res, self.exp_bad_oc, atol=1e-4)

    def test_bad_indata(self):
        """Test the CloudCompositorWithoutCloudfree composite generation without status."""
        data = xr.DataArray(np.array([[4, 3, 2], [2, 3, 4], [255, 7, 255]], dtype=np.uint8),
                            dims=["y", "x"],
                            attrs={"_FillValue": 255,
                                   "scaled_FillValue": 255})
        np.testing.assert_raises(ValueError, self.colormap_composite, [data])


class TestCloudCompositorCommonMask:
    """Test the CloudCompositorCommonMask."""

    def setup_method(self):
        """Set up the test case."""
        from satpy.composites.cloud_products import CloudCompositorCommonMask

        self.exp_a = np.array([[4, 3, 2],
                               [2, 3, 655350],
                               [np.nan, np.nan, np.nan]])
        self.exp_b = np.array([[4, 3, 2],
                               [2, 3, 255],
                               [np.nan, np.nan, np.nan]])
        self.colormap_composite = CloudCompositorCommonMask("test_cmap_compositor")

    def test_call_numpy(self):
        """Test the CloudCompositorCommonMask with numpy."""
        mask = xr.DataArray(np.array([[0, 0, 0], [1, 1, 1], [255, 255, 255]]), dims=["y", "x"],
                            attrs={"_FillValue": 255})
        data = xr.DataArray(np.array([[4, 3, 2], [2, 3, np.nan], [np.nan, np.nan, np.nan]], dtype=np.float32),
                            dims=["y", "x"],
                            attrs={"_FillValue": 65535,
                                   "scaled_FillValue": 655350})
        res = self.colormap_composite([data, mask])
        np.testing.assert_allclose(res, self.exp_a, atol=1e-4)

    def test_call_dask(self):
        """Test the CloudCompositorCommonMask with dask."""
        mask = xr.DataArray(da.from_array(np.array([[0, 0, 0], [1, 1, 1], [255, 255, 255]])), dims=["y", "x"],
                            attrs={"_FillValue": 255})
        data = xr.DataArray(da.from_array(np.array([[4, 3, 2], [2, 3, 255], [255, 255, 255]], dtype=np.int16)),
                            dims=["y", "x"],
                            attrs={"_FillValue": 255,
                                   "scaled_FillValue": 255})
        res = self.colormap_composite([data, mask])
        np.testing.assert_allclose(res, self.exp_b, atol=1e-4)

    def test_bad_call(self):
        """Test the CloudCompositorCommonMask without mask."""
        data = xr.DataArray(np.array([[4, 3, 2], [2, 3, 255], [255, 255, 255]], dtype=np.int16),
                            dims=["y", "x"],
                            attrs={"_FillValue": 255,
                                   "scaled_FillValue": 255})
        np.testing.assert_raises(ValueError, self.colormap_composite, [data])


class TestPrecipCloudsCompositor(unittest.TestCase):
    """Test the PrecipClouds compositor."""

    def test_call(self):
        """Test the precip composite generation."""
        from satpy.composites.cloud_products import PrecipCloudsRGB
        colormap_compositor = PrecipCloudsRGB("test_precip_compositor")

        data_light = xr.DataArray(np.array([[80, 70, 60, 0], [20, 30, 40, 255]], dtype=np.uint8),
                                  dims=["y", "x"], attrs={"_FillValue": 255})
        data_moderate = xr.DataArray(np.array([[60, 50, 40, 0], [20, 30, 40, 255]], dtype=np.uint8),
                                     dims=["y", "x"], attrs={"_FillValue": 255})
        data_intense = xr.DataArray(np.array([[40, 30, 20, 0], [20, 30, 40, 255]], dtype=np.uint8),
                                    dims=["y", "x"], attrs={"_FillValue": 255})
        data_flags = xr.DataArray(np.array([[0, 0, 4, 0], [0, 0, 0, 0]], dtype=np.uint8),
                                  dims=["y", "x"])
        res = colormap_compositor([data_light, data_moderate, data_intense, data_flags])

        exp = np.array([[[0.24313725, 0.18235294, 0.12156863, np.nan],
                         [0.12156863, 0.18235294, 0.24313725, np.nan]],
                        [[0.62184874, 0.51820728, 0.41456583, np.nan],
                         [0.20728291, 0.31092437, 0.41456583, np.nan]],
                        [[0.82913165, 0.7254902, 0.62184874, np.nan],
                         [0.20728291, 0.31092437, 0.41456583, np.nan]]])

        np.testing.assert_allclose(res, exp)
